#include "korg.hpp"
#include "korgtext.hpp"
#include <string.h>

char* styname;
FILE* styf;
STYHEAD styhead;
STYDATA stydata;
STYLINK stylink;
int stylecount = 0;
int userstyle = -1;
KORGNAME name;
long styofstab = -1;
long curstyle = -1, nextstyle = -1;
long firstpattern = 0x7fffffffl;

const char* getstyle(int i)
{
char *styleplace = "Ux";

  styleplace[1] = (i % 8)+'1';
  return (const char*)styleplace;
}

void dumpstylenames()
{
int count;
KORGNAME s;

  count = (int)(styhead.stylenames.len / KORGNAME_SIZE);

  printf("%d STYLE NAMES: \n", count);

  fseek(styf, styhead.stylenames.adr, SEEK_SET);
  for (userstyle = 0; userstyle < count && userstyle < 4; userstyle++)
  {
    if (!read_korgname(styf, s))
      return;
    printf("  %4s  %10s", getstyle(userstyle), s);
    if (userstyle % 4 == 3)
      printf("\n");
  }
  printf("\n");
}


void stylehdr(int i, char* s)
{
//  printf("\nSTYLE %10.10s %s %s\n", stydata.name, getstyle(i), s);
}

int getlink(int i)
{
long oldpos = ftell(styf);
int ret;

  fseek(styf, styhead.styletable.adr + sizeof(stylink) * i, SEEK_SET);
  ret = KORGREAD(styf, stylink);
  fseek(styf, oldpos, SEEK_SET);
  return ret;
}

void dumpstyletable()
{
int i;

  stylecount = (int)(styhead.styletable.len / sizeof(STYLINK));

  printf("%d STYLE TABLE: \n", stylecount);
  if (sizeof(STYLINK) != 6)
    putchar(7);


  for (i = 0; i < stylecount; i++)
  {
    if (!getlink(i))
      return;

    printf("STYLE U%d: ", i+1);
    dump(stylink.unknown, 4);
    printf("len = %u\n", stylink.len);
    printf("\n");
  }
}

const char* variation2str(int i)
{
static char varstr[20];
int v, cv;

  if (i >= 0 && i < 24)
  {
    v = i / 6;
    cv = i % 6;
  }
  else if (i < 216)
  {
    v = (i - 24) / 2 + 4;
    cv = (i - 24) % 2;
  }
  else
    return "???";

  sprintf(varstr, "%-7s CV%d",
    styelement2str(v),
    cv+1);
  return varstr;
}

const char* chordkey2str(int key)
{
static char s[4];
char* p = s;

  *p++ = key2char(key / 3);
  if (key % 3 == 1)
    *p++ = 'm';
  else if (key % 3 == 2)
    *p++ = '#';
  *p = 0;
  return (const char*)s;
}



void dumptrack(char* track, TRACKINFO& trinfo)
{
  printf("%s: instr=%s  volume=%03d PAN=%s\n",
    track,
    place2str(trinfo.instr),
    trinfo.volume,
    pan2str(trinfo.panpot));
}

void dumpchordvar(char* s, CHORDVAR& chvar)
{
  printf("%s:  ", s);
  for (int i = 0; i <sizeof(chvar); i++)
    printf("%d ", chvar[i]+1);
  printf("\n");
}

void dumppatparam(int pnr, PATPARAM& param)
{
    printf("Pattern%02d:   Tact=%s  Length=%02d\n",
      pnr, tact2str(param.tact), param.len);
}

void dumpallpatparam(ALLPATPARAM & allp)
{
  for (int i = 0; i < 100; i++)
    dumppatparam(i, allp[i]);
}

void dumpevent(EVENT &ev)
{
  switch(ev.general[1] & 0xf0)
  {
  case 0x10:
    printf("  %5s PITCH %03d", loc2str(ev.pitch.loc), (ev.pitch.v_hi << 7) + ev.pitch.v_lo);
    break;
  case 0x20:
    printf("  %5s AFTT %03d", loc2str(ev.aftt.loc), ev.aftt.v);
    break;
  case 0x30:
    printf("  %5s PROG %s", loc2str(ev.prog.loc), place2str(ev.prog.prog));
    break;
  case 0x40:
    printf("  %5s CTRL %03d %03d",
      loc2str(ev.ctrl.loc),
      ev.ctrl.control, ev.ctrl.value);
    break;
  case 0x50:
    printf("  %5s POLY %03d %03d",
      loc2str(ev.poly.loc),
      ev.poly.key, ev.poly.value);
    break;
  case 0x60:
    printf("  MEASURE %d at beat of %s ", ev.beat.index+1, tact2str(ev.beat.tact));
    if (ev.beat.pattern_option & 2)
      printf("* PATTERN %02d *", ev.beat.pattern);
    break;
  case 0x70:
    printf("** END TRACK ** %d", ev.end.bar);
    break;
  default:
    if (ev.note.cmd == 1)
    {
      printf("  %5s %-3s   V%03d",
	loc2str(ev.note.loc),
	note2str(ev.note.key, -1),
	ev.note.vel * 2);
      printf("  %s", len2str(ev.note.len));
    }
    else
    {
      printf("  OTHER %02X %02X %02X %02X",
	ev.general[0], ev.general[1],
	ev.general[2], ev.general[3]);
    }
    break;
  }
  printf("\n");
}

void dumppattern( long len)
{
EVENT ev;

  if (len == 0)
    printf(" (empty)");
  else
  {

    while (len > 0)
    {
      if (!KORGREAD(styf, ev))
	return;
      len -= 4;

      dumpevent(ev);
    }

//    fseek(styf, len, SEEK_CUR);
    if (len % 4)
      putchar(7);
  }
  printf("\n");
}

void dumpchordseq(STYDATA& sty)
{
long len;

  for (int i = 0; i < 216; i++)
  {
    len = sty.seqofs[i+1 + CHORDOFS] - sty.seqofs[i + CHORDOFS];
    if (len > 0)
    {
      fseek(styf, (sty.seqofs[i + CHORDOFS] + curstyle), SEEK_SET);
      printf("%04lX: %s CHORDSEQ %02d: %s %s\n", ftell(styf), getstyle(userstyle),
	i, variation2str(i/6), track2str(i % 6));
      dumppattern(len);
    }
  }
}

void dumppatternseq(STYDATA& sty)
{
long len;

  for (int i = 0; i < 100; i++)
  {
    len = sty.seqofs[i+1 + PATTERNOFS] - sty.seqofs[i + PATTERNOFS];
    if (len > 0)
    {
      fseek(styf, (sty.seqofs[i + PATTERNOFS] + curstyle), SEEK_SET);
      printf("%04lX: %s PATTERN %02d:\n", ftell(styf), getstyle(userstyle), i);
      dumppattern(len);
    }
  }
}

void dumppatterns(STYDATA& sty)
{
  dumpchordseq(sty);
  dumppatternseq(sty);
}

void dumpchordparam(int i, CHORDPARAM & cpar)
{
  printf("  %s Length=%02d Key=%s\n",
    variation2str(i),
    cpar.len,
    chordkey2str(cpar.key)
    );
}

void dumpchordparams(ALLCHORDPARAM & cpar)
{
  printf("CHORDVAR PARAMS:\n");
  for (int i = 0; i < 36; i++)
    dumpchordparam(i, cpar[i]);
}

void dumpstyledata(int stynr)
{
STYDATA sty;

  if (!KORGREAD(styf, sty))
    return;
  printf("STYLE DATA %d: %10.10s\n", stynr, name);
  switch(sty.param.author)
  {
  case USERSTYLE:  printf("USER CREATED STYLE\n"); break;
  case BUILTINSTYLE:  printf("BUILT-IN STYLE \n"); break;
  case CARDSTYLE:  printf("CARD STYLE\n"); break;
  default: printf("author unknown: %d\n", sty.param.author); break;
  }
  printf("TEMPO: %d\n", 30 + sty.param.tempo);
  printf("TACT: %s   RESOLUTION: %s\n", tact2str(sty.param.tact), resol2str(sty.param.tact));
  printf("NOTE RETRIGGER: ");
  printf("BASS=%s ", onoff2str(sty.param.retrig_bass));
  printf("ACC1=%s ", onoff2str(sty.param.retrig_acc1));
  printf("ACC2=%s ", onoff2str(sty.param.retrig_acc2));
  printf("ACC3=%s\n", onoff2str(sty.param.retrig_acc3));
  printf("TENSION: ");
  printf("ACC1=%s ", onoff2str(sty.param.tension_acc1));
  printf("ACC2=%s ", onoff2str(sty.param.tension_acc2));
  printf("ACC3=%s\n", onoff2str(sty.param.tension_acc3));
  printf("WRAP AROUND POINTS:  BASS=%02d  ACC1=%02d ACC2=%02d ACC3=%02d\n",
    sty.param.wrap.bass+1,
    sty.param.wrap.acc[0]+1,
    sty.param.wrap.acc[1]+1,
    sty.param.wrap.acc[2]+1);
//  dump(sty.reserve, 20); printf("\n");
  for (int i = 0; i < 6; i++)
    dumptrack((char*)track2str(i), sty.trinfo[i]);

  dumpchordparams(sty.chordparam);

  printf("\n");

  dumpchordvar("VARIATION 1", sty.chvar.var[0]);
  dumpchordvar("VARIATION 2", sty.chvar.var[1]);
  dumpchordvar("VARIATION 3", sty.chvar.var[2]);
  dumpchordvar("VARIATION 4", sty.chvar.var[3]);
  dumpchordvar("    INTRO 1", sty.chvar.intro[0]);
  dumpchordvar("    INTRO 2", sty.chvar.intro[1]);
  dumpchordvar("   ENDING 1", sty.chvar.ending[0]);
  dumpchordvar("   ENDING 2", sty.chvar.ending[1]);
  dumpchordvar("     FILL 1", sty.chvar.fill[0]);
  dumpchordvar("     FILL 2", sty.chvar.fill[1]);
  dumpallpatparam(sty.patparam);
  dumppatterns(sty);
}

void dumpsty()
{
  if (!styheader(styf, styhead))
  {
    fprintf(stderr, "%s is not a KORG sty file\n", styname);
    return;
  }
  printf("\nKORG STY file %s\n\n", styname);

  dumpstylenames();

  dumpstyletable();

  curstyle = -1;
  nextstyle = styhead.styletable.adr + styhead.styletable.len;
  for (userstyle = 0; userstyle < stylecount; userstyle++)
  {
    fseek(styf, nextstyle, SEEK_SET);
    if (!getlink(userstyle))
      return;
    curstyle = nextstyle;
    nextstyle += stylink.len;

    dumpstyledata(userstyle);
    if (nextstyle != ftell(styf))
    {
      if (nextstyle > ftell(styf))
	 dumpf(styf, nextstyle - ftell(styf));
      printf("KORREKTUR %ld bytes\n", nextstyle - ftell(styf));
    }
  }
}

void main(int argc, char**argv)
{
  argc--; argv++;

  if (argc == 0)
    return;

  styname = *argv++; argc--;

  styf = fopen(styname, "rb");
  if (!styf)
  {
    perror(styname);
    return;
  }

  dumpsty();

  fclose(styf);
}
