#ifndef __PCG__
#define __PCG__
#include <stdio.h>

#define KORGREAD(file, obj) \
   ((fread(&(obj), sizeof(obj), 1, file) != 1) ? 0 : 1)


#define KORGNAME_SIZE  10
typedef char KORGNAME[KORGNAME_SIZE+1];

#define CHORDOFS      0
#define PATTERNOFS  216

int read_korgname(FILE* f, char *s);

typedef struct
{
  long adr;
  long len;
} block;

typedef struct
{
  char korg[4];
  char unknown[12];
  block b1;
  block prognames;
  block globals;
  block drumkits;
  block b5;
  block progdata;
  block b7;
  block b8;
  block b9;
} PCGHEAD;

typedef unsigned char ubyte;
typedef signed char sbyte;

typedef unsigned short uword;

typedef struct
{
  sbyte mastertune;
  ubyte keytranspose;
  ubyte damperpol;
  ubyte pedal1;
  ubyte pedal2;
  ubyte unknown5;
  ubyte globalchannel;
  ubyte unknown7;
  ubyte unknown8;
  ubyte unknown9;
  ubyte mainscaletype;
  ubyte mainscalekey;
  ubyte userscale[12];
  ubyte unknown18;
  ubyte velcurve;
  ubyte aftertouchcurve;
  ubyte unknown1b;
  ubyte unknown1c;
  ubyte subscaletype;
  ubyte subscalekey;
  ubyte unknown1f;
  ubyte drumchannel;
  ubyte percchannel;
  ubyte basschannel;
  ubyte acc1channel;
  ubyte acc2channel;
  ubyte acc3channel;
  ubyte switchA;
  ubyte switchB;
  ubyte switchC;
  ubyte switchD;
  ubyte switchE;
} PCGGLOBAL;

typedef struct
{
  int drum;
  ubyte key;
  ubyte panlevel : 5;
  ubyte group : 3;
  char d1[2];
  sbyte level;
  char d2;
  sbyte decay;
  ubyte effect_sendlevel; // C-D
  char d3[12];
} PCGDRUM;

#define DRUMSPERKIT  (60)

typedef struct
{
  ubyte key;
  sbyte kbd_trk_amount;
  sbyte vel_sense_amount;
  ubyte kbd_trk_time;
  ubyte vel_sense_time;
  ubyte attacktime;
  sbyte attacklevel;
  ubyte decay_time;
  sbyte break_point;
  ubyte slope_time;
  sbyte sustain_level;
  ubyte release_time;
} VDA;

typedef struct
{
  ubyte freq;
  ubyte delay;
  ubyte fade_in;
  sbyte intens;
  sbyte kbd_tracking;
  ubyte joystick;
  ubyte aftertouch;
  ubyte freqctrl;
} PITCH;

typedef struct
{
  PITCH pitch;
  ubyte cutoff;
  ubyte pivot_key;  // C-1 is 0 .. G9
  sbyte kbd_trk_amount; // -99..+99
  ubyte intensity;
  ubyte kbd_trk_time;
  ubyte vel_sense_time;
  sbyte vel_sense_amount;
  ubyte attacktime;
  sbyte attacklevel;
  ubyte decay_time;
  sbyte break_point;
  ubyte slope_time;
  sbyte sustain_level;
  ubyte release_time;
  sbyte release_level;
  ubyte level;
  VDA vda;
  ubyte kbd_timeflags;
  ubyte vel_timeflags;
  ubyte vda_kbd_timeflags;
  ubyte vda_vel_timeflags;
  ubyte effect_send_level;
  ubyte kbdcolor;
  sbyte velcolor;
  ubyte mode;
  ubyte padpot;
} VDF;

typedef struct
{
  char name[KORGNAME_SIZE];
  char doubleoscillator;
  char monophonic : 1;
  char hold : 1;
  int  osc1_multisound;
  char row1[1];
  int  osc2_multisound;

  // 1 byte from int!
  char row2a[1];
  sbyte intervall;
  sbyte detune;
  ubyte delay;
  sbyte startlevel;
  ubyte attacktime;
  sbyte attacklevel;
  ubyte decaytime;
  ubyte releasetime;
  sbyte releaselevel;
  sbyte timevelsense;
  sbyte velsense;
  ubyte vdfmg_waveform : 5;
  ubyte vdfmg_osc : 2;
  ubyte vdfmg_sync : 1;
  ubyte vdfmg_freq;
  ubyte vdfmg_delay;

  ubyte vdfmg_intens;
  sbyte bend_aftertouch;
  sbyte cutoff_aftertouch;
  ubyte vdfmg_aftertouch;
  sbyte level_aftertouch;
  sbyte bend_joystick;
  sbyte cutoff_joystick;
  ubyte vdfmg_joystick;
  char row3b[2];

  VDF  vdf1;

  char row6c[2];

  VDF  vdf2;

  char row9c[1];
  ubyte fx1;
  ubyte fx2;
  char row9d[6];
  ubyte fxflags;

  ubyte fx_vib_depth;
  ubyte fx_accel;
  ubyte fx_slowspeed;
  ubyte fx_fastspeed;
  char row10a[4];
  ubyte fx1_dmod_source;
  sbyte fx1_dmod_amount;
  ubyte fx1_delay_time;
  ubyte fx_mod_speed;
  ubyte fx_mod_depth;
  sbyte fx1_feedback;
  ubyte fx2_delay_time_half;
  sbyte fx2_feedback;

  char row11[2];
  ubyte fx2_dmod_source;
  sbyte fx2_dmod_amount;
} PCGDATA;

int pcgheader(FILE* f, PCGHEAD &head);

typedef struct
{
  char korg[4];
  char unknown[12];
  block stylenames;
  block styletable;
} STYHEAD;

typedef struct
{
  char unknown[4];
  unsigned len;
} STYLINK;

typedef struct
{
  int instr;
  ubyte volume;
  ubyte panpot;
} TRACKINFO;

typedef TRACKINFO ALLTRACKINFO[6];

typedef ubyte CHORDVAR[20];

typedef struct
{
  CHORDVAR var[4], intro[2], ending[2], fill[2];
} ALLCHORDVAR;

typedef struct
{
  ubyte bass, acc[3];
} WRAPAROUND;

// style author:
#define USERSTYLE    0
#define BUILTINSTYLE 1
#define CARDSTYLE    2

typedef struct
{
  ubyte author;
  ubyte tempo;
  ubyte tact;
  ubyte f1 : 2;
  ubyte retrig_bass : 1;
  ubyte retrig_acc1 : 1;
  ubyte retrig_acc2 : 1;
  ubyte retrig_acc3 : 1;
  ubyte f2 : 2;
  WRAPAROUND wrap;
  ubyte t1 : 3;
  ubyte tension_acc1 : 1;
  ubyte tension_acc2 : 1;
  ubyte tension_acc3 : 1;
  ubyte t2 : 2;
} STYPARAM;

typedef struct
{
  ubyte tact;
  ubyte len;
} PATPARAM;

typedef PATPARAM ALLPATPARAM[100];

typedef struct
{
  uword loc : 9;
  uword vel : 6;
  uword cmd : 1;

  uword len : 9;
  uword key : 7;
} NOTEEVENT;

// pattern_option
#define DONT_USE_PATTERN  0
#define PATTERN_CONTINUE  2
#define PATTERN_START     3

typedef struct
{

  uword index : 10;
  ubyte cmd : 6;
  ubyte pattern;
  ubyte tact : 6;
  ubyte pattern_option : 2;
} BEATEVENT;

typedef struct
{
  uword loc : 9;
  uword cmd : 7;
  ubyte control;
  ubyte value;
} CTRLEVENT;

typedef struct
{
  uword loc : 9;
  uword cmd : 7;
  uword v;
} AFTTEVENT;

typedef struct
{
  uword loc : 9;
  uword cmd : 7;
  ubyte v_lo;
  ubyte v_hi;
} PITCHEVENT;

typedef struct
{
  uword loc : 9;
  uword cmd : 7;
  uword prog;
} PROGEVENT;

typedef struct
{
  uword loc : 9;
  uword cmd : 7;
  ubyte key;
  ubyte value;
} POLYEVENT;

typedef struct
{
  uword loc : 9;
  uword cmd : 7;
  uword bar;
} ENDEVENT;

typedef union
{
  NOTEEVENT note;
  BEATEVENT beat;
  CTRLEVENT ctrl;
  AFTTEVENT aftt;
  PITCHEVENT pitch;
  PROGEVENT prog;
  POLYEVENT poly;
  ENDEVENT  end;
  ubyte general[4];
} EVENT;

typedef struct
{
  ubyte key;
  ubyte len;
} CHORDPARAM;

typedef CHORDPARAM ALLCHORDPARAM[36];

typedef struct
{
  char name[KORGNAME_SIZE];
  STYPARAM param;
  char reserve[19];
  ALLTRACKINFO trinfo;
  ALLCHORDPARAM chordparam;
  ALLCHORDVAR chvar;
  ALLPATPARAM patparam;
  unsigned int seqofs[317];
} STYDATA;

int styheader(FILE* f, STYHEAD & styhead);

typedef struct
{
  int   prog;
  ubyte volume;
  ubyte panpot;
  ubyte effectsendlevel;
  sbyte octave;
  ubyte unknown2[1];
  sbyte wraparound;
} TRKPARAM;

typedef struct
{
  ubyte maxvel, minvel;
} VELWINDOW;

typedef struct
{
  char name[10];
  ubyte style;
  char x1[2];
  ubyte drum;
  ubyte tempo;
  char x5[1];
  ubyte split;
  char x2[5];
  ubyte defdrum[4];
  ubyte kickdesignation;
  ubyte snaredesignation;
  char x3[2];
  ubyte varchange1;
  char x7[1];
  ubyte varchange2;
  char x6[1];
  TRKPARAM trkparam[6];
  TRKPARAM kbdparam[2];
  VELWINDOW kbdwin[2];
  char x4[29];
} ARRDATA;

typedef struct
{
  char korg[4];
  char unknown[12];
  block arrnames;
  block arrdata;
} ARRHEAD;

typedef struct
{
  uword prog;
  sbyte octave;
} KBDATTR;

typedef struct
{
  uword prog;
  ubyte vol;
  ubyte pan;
  ubyte cd_sendlevel;
  ubyte trackstat;
  ubyte bendrange;
  sbyte transpose;
  ubyte detune;
  ubyte protect;
  ubyte channel;
  ubyte vel_wintop;
  ubyte vel_winbottom;
  ubyte key_wintop;
  ubyte key_winbottom;
} TRACK_CTRLDATA;

typedef struct
{
  char c[29];
} EFFECTPARAM;

typedef struct
{
  unsigned adr[10][17];
  unsigned endadr;
} BSQDATAADR;

typedef struct
{
  char name[10];
  uword arrno;
  uword styno;
  uword variation;
  ubyte tempo;
  ubyte keybassign;
  ubyte chord_scan_low : 1;
  ubyte chord_scan_high : 1;
  ubyte bass_invers : 1;
  ubyte chord_hold : 1;
  ubyte chord_latch : 1;
  ubyte chord_reserved : 3;
  KBDATTR kbdattr[2];
  ubyte kbd_trkstatus;
  ubyte cntrl_trkstatus;
  ubyte chord_trkstatus;
  ubyte autotempo;
  ubyte beat;
  ubyte splitpoint;
  sbyte transpose;
  ubyte sw_dynvel : 1;
  ubyte sw_reserv1 : 7;
  ubyte sw_reserv2[11];
  ubyte next_bsq;
  ubyte autostart;
  TRACK_CTRLDATA track[8];
  EFFECTPARAM effects;
} BSQDATA;

int arrheader(FILE* f, ARRHEAD & styhead);

typedef struct
{
  char korg[4];
  char unknown[12];
} BSQHEAD;

typedef struct
{
  uword time:9;
  uword chord:5;
  uword cmd:2;
  uword root : 4;
  uword bass : 4;
  uword tensionnote : 8;
} CHORDEVENT;

typedef struct
{
  uword time:9;
  uword control : 5;
  uword cmd:2;
  ubyte v1, v2;
} CONTROLEVENT;

typedef union
{
  CHORDEVENT chord;
  CONTROLEVENT control;
  BEATEVENT beat;
  ENDEVENT end;
  CTRLEVENT ctrl;
  PROGEVENT prog;
  NOTEEVENT note;
  ubyte general[4];
} BSQEVENT;

int bsqheader(FILE* f, BSQHEAD & head);

typedef struct
{
  char korg[4];
  char unknown[12];
} SNGHEAD;

typedef struct
{
  ubyte syncsrc : 2; // 0=OFF, 1=INT, 2=EXT, 3=BOTH
  ubyte mute : 1;    // 0=PLAY, 1=MUTE
  ubyte unknown : 5;
} SNGSTATUS;

typedef struct
{
  ubyte prog;
  ubyte outlevel;
  sbyte transpose;
  sbyte detune;
  ubyte pan;
  ubyte sendlevel;
  ubyte key_wintop;
  ubyte key_winbottom;
  ubyte vel_wintop;
  ubyte vel_winbottom;
  ubyte ctrlfilter;
  ubyte channel;
} TRACKCTRL;

typedef struct
{
  ubyte channel[16];
  SNGSTATUS status[16];
  ubyte bendrange[16];
  ubyte beat;
  ubyte tempo;
  ubyte protect1; // bit track 1-8
  ubyte protect2; // bit track 9-16
  ubyte nextsong; // bit 0-6: song 0-9  7F=OFF  bit7: autostart
  char  songtitle[10];
  ubyte reserved1; // reserved 0
  EFFECTPARAM effects;
  TRACKCTRL trackctrl[16];
  ubyte reserved2[6];  // 0
  ubyte metron_level;
  ubyte metron_pan;
  ubyte metron_leadin;
  ubyte tempotrack;
  ubyte reserved3[1];
} SNGCTRL;

typedef struct
{
  ubyte beat;
  ubyte len;
} PATCTRL;

typedef struct
{
  unsigned sngadr[10][17];
  unsigned patadr[100];
  unsigned endadr;
} SNGADR;

typedef struct
{
  SNGCTRL sngctrl[10];
  PATCTRL patctrl[100];
  SNGADR  adr;
} SNG;

int sngheader(FILE* f, SNGHEAD & head);
void sngdefheader(SNGHEAD & head);
void sngdefcontrol(SNGCTRL& ctrl);
void sngdefpatctrl(PATCTRL & pat);
void sngdefsong(SNG & sng);

#endif