#include "korgtext.hpp"
#include <string.h>

char key2char(int note)
{
  switch(note)
  {
  case 0:
  case 1: return 'C';
  case 2:
  case 3: return 'D';
  case 4: return 'E';
  case 5:
  case 6: return 'F';
  case 7:
  case 8: return 'G';
  case 9:
  case 10: return 'A';
  case 11: return 'H';
  }
  return '?';
}

const char* note2str(int note, int baseoct)
{
int oct = note / 12 + baseoct ;
static char* nstr = "C#-4";
char* p = nstr;

  note %= 12;

  *p++ = key2char(note);
  switch(note)
  {
  case 1:
  case 3:
  case 6:
  case 8:
  case 10:  *p++ = '#'; break;
  }
  sprintf(p, "%d", oct);
  return (const char*)nstr;
}

const char* efflev2str(int c, int invers)
{
static char* efflev = "C=?  D=?";

  efflev[invers ? 7 : 2] = (c >> 4) + '0';
  efflev[invers ? 2 : 7] = (c & 15) + '0';
  return (const char*)efflev;
}

const char*multisound2str(int i)
{
  switch(i)
  {
  case 0:  return "A.Piano 1";
  case 11: return "PianoPad 1";
  case 16: return "HarpsicdLP";
  case 222:  return "Tap 1";
  default: return "Multi???";
  }
}

const char* pan2str(int i)
{
static char* pans = "Lxx";

  if (i >= 0 && i <= 14)
    sprintf(pans, "L%02d", 15-i);
  else if (i == 15)
    return "CNT";
  else if (i == 31)
    return "PRG";
  else if (i > 15)
    sprintf(pans, "R%02d", i - 15);
  else if (i == 255)
    return "OFF";
  else
    return "???";
  return (const char*) pans;
}

const char* mode2str(int mode, int isvda)
{
int m;

  if (isvda)
    m = (mode >> 4) & 3;
  else
    m = mode & 3;
  switch(m)
  {
  case 0:  return "OFF";
  case 1:  return "LOW";
  case 2:  return "HIGH";
  case 3:  return "ALL";
  default: return "MODE???";
  }
}

char pm2char(unsigned char b, int which)
{
int v;

  v = ((b & (1 << (which+4))) ? 2 : 0) + ((b & (1 << which)) ? 1 : 0);
  switch(v)
  {
  case 0:  return '0';
  case 1: return '+';
  case 3: return '-';
  default: return '?';
  }
}

const char*onoff2str(int n)
{
  return n ? "ON " : "OFF";
}

const char*sync2str(int n)
{
  return n ? "SYNC" : "OFF ";
}

const char* waveform2str(int w)
{
  switch(w)
  {
  case 0: return "TRIANGLE";
  case 1: return "SAW UP";
  case 2: return "SAW DOWN";
  case 3: return "RANDOM";
  default: return "WAVE???";
  }
}

const char* effect2str(int e)
{
  switch(e)
  {
  case 1: return "Hall";
  case 17: return "MultiTap Dly2";
  case 28: return "Exiter";
  case 34: return "Rotary Speaker";
  case 38: return "Chorus-Delay";
  default: return "FX???";
  }
}

const char* dmod2str(int d)
{
  switch(d)
  {
  case 0: return "NONE";
  case 1: return "JS UP";
  case 2: return "JS DWN";
  case 3: return "AFTT";
  case 4: return "PEDAL1";
  case 5: return "PEDAL2";
  case 6: return "VDA EG";
  default: return "MOD???";
  }
}

const char* group2str(int grp)
{
static char* excl = "EX?";

  if (grp == 0)
    return "---";
  if (grp >= 1 && grp <= 6)
  {
    excl[2] = grp + '0';
    return (const char*)excl;
  }
  if (grp == 7)
    return "SLF";
  return "???";
}

const char* trackstat2str(int stat)
{
  if (stat & 4)
    return "MUTE";
  switch(stat & 3)
  {
  case 0: return "OFF";
  case 1: return "INT";
  case 2: return "EXT";
  case 3: return "BOTH";
  }
  return "???";
}

const char* tact2str(int tact)
{
static char t[8] = "?/?";

  if (tact >= 0x10 && tact <= 0x18)
    sprintf(t, "%d/4", tact-0x0f);
  else if (tact >= 0x20 && tact <= 0x2f)
    sprintf(t, "%d/8", tact-0x1f);
  else if (tact >= 0x30 && tact <= 0x3f)
    sprintf(t, "%d/16", tact-0x2f);
  return (const char*)t;
}

const char* resol2str(int tact)
{
  return (tact & 0x80) ? "Low" : "High";
}

const char* place2str(int i)
{
char *progplace = "Dxx";

  switch(i & 0x340)
  {
  case 0:     progplace[0] = 'A'; break;
  case 0x40:  progplace[0] = 'B'; break;
  case 0x100: progplace[0] = 'C'; break;
  case 0x140: progplace[0] = 'D'; break;
  case 0x200: progplace[0] = 'D'; progplace[1] = 'r'; break;
  default:    progplace[0] = '?'; break;
  }

  if ((i & 0x200) == 0)
    progplace[1] = ((i / 8) % 8)+'1';
  progplace[2] = (i % 8)+'1';
  return (const char*)progplace;
}

const char* track2str(int i)
{
  switch(i)
  {
  case 0: return "DRUM";
  case 1: return "PERC";
  case 2: return "BASS";
  case 3: return "ACC1";
  case 4: return "ACC2";
  case 5: return "ACC3";
  default: return "Tr??";
  }
}

const char* style2str(int i)
{
static char s[4];

  if (i >= 0 && i < 48)
  {
    s[0] = 'P';
    s[1] = '1' + i / 8;
    s[2] = '1' + i % 8;
    s[3] = 0;
  }
  else if (i < 52)
  {
    s[0] = 'U';
    s[1] = '1' + (i - 48);
    s[2] = 0;
  }
  else if (i < 56)
  {
    s[0] = 'C';
    s[1] = '1' + (i - 52);
    s[2] = 0;
  }
  else
    return "style??";
  return (const char*)s;
}

const char* kbdassign2str(int i)
{
  switch(i)
  {
  case 0:  return "SINGLE";
  case 1:  return "LAYER";
  case 2:  return "SPLIT";
  case 3:  return "M. DRUMS";
  default: return "???";
  }
}

void dump( char* c, unsigned len )
{
  for (int i = 0; i < len; i++, c++)
    printf("%02X ", (unsigned char)*c);
}

void dumpf( FILE* f, unsigned len )
{
  for (unsigned i = 0; i < len; i++)
  {
    if (i % 16 == 0)
      printf("%04lX:", ftell(f));

    printf(" %02X", (unsigned char)fgetc(f));

    if (i % 16 == 15 && i < len)
      printf("\n");
  }
}

const char* time2str(unsigned n, int isloc)
{
static char tm[6];

  if ((!isloc && n == 0xfe) || n == 0x1fe)
    return "TIE";

  sprintf(tm, "%d.%02d", n / 96 + (isloc ? 1 : 0), n % 96);
  return (const char*)tm;
}

const char* loc2str(unsigned n)
{
  return time2str(n, 1);
}

const char* len2str(unsigned n)
{
  return time2str(n, 0);
}

const char* styelement2str(int i)
{
  switch(i)
  {
  case 0: return "Var1";
  case 1: return "Var2";
  case 2: return "Var3";
  case 3: return "Var4";
  case 4: return "Intro1";
  case 5: return "Intro2";
  case 6: return "Ending1";
  case 7: return "Ending2";
  case 8: return "Fill1";
  case 9: return "Fill2";
  default: return "styelement???";
  }
}

const char* chordtype2str(int c)
{
  switch(c)
  {
  case 0: return "-";
  case 1: return "Maj";
  case 2: return "6";
  case 3: return "M7";
  case 4: return "M7b5";
  case 5: return "sus4";
  case 6: return "sus2";
  case 7: return "M7sus4";
  case 8: return "m";
  case 9: return "m6";
  case 10: return "m7";
  case 11: return "m7b5";
  case 12: return "mM7";
  case 13: return "7";
  case 14: return "7b5";
  case 15: return "7sus4";
  case 16: return "dim";
  case 17: return "dimM7";
  case 18: return "aug";
  case 19: return "aug7";
  case 20: return "augM7";
  default: return "Chord???";
  }
}

const char* tension2str(int c)
{
static char t[30];

  t[0] = 0;
  if (c & 0x01) strcat(t, "b9 ");
  if (c & 0x02) strcat(t, "9 ");
  if (c & 0x04) strcat(t, "#9 ");
  if (c & 0x08) strcat(t, "11 ");
  if (c & 0x10) strcat(t, "#11 ");
  if (c & 0x20) strcat(t, "b13 ");
  if (c & 0x40) strcat(t, "13 ");
  return (const char*)t;
}

const char* chord2str(int c, int tension, int bass, int root)
{
static char ch[40];

  sprintf(ch, "%c %s %s /%c",
    key2char(bass),
    chordtype2str(c),
    tension2str(tension),
    key2char(root));
  return (const char*)ch;
}

const char* arr2str(int i)
{
static char a[3];

  a[0] = i / 8 + '1';
  a[1] = i % 8 + '1';
  a[2] = 0;
  return (const char*)a;
}

const char* chordscan2str(int i)
{
  switch(i)
  {
  case 0: return "OFF";
  case 1: return "LOWER";
  case 2: return "UPPER";
  case 3: return "FULL";
  default: return "scan???";
  }
}

const char* place2str(int prg, int bank)
{
  return place2str((bank << 8) + prg);
}

const char* bseqtrack(int i)
{
  switch(i)
  {
  case 0:  return "KBD TRACK";
  case 1:  return "CTRL TRACK";
  case 2:  return "CHORD TRACK";
  case 8:  return "EXTRA TRACK 1";
  case 9:  return "EXTRA TRACK 2";
  case 10:  return "EXTRA TRACK 3";
  case 11:  return "EXTRA TRACK 4";
  case 12:  return "EXTRA TRACK 5";
  case 13:  return "EXTRA TRACK 6";
  case 14:  return "EXTRA TRACK 7";
  case 15:  return "EXTRA TRACK 8";
  case 16:  return "TEMPO TRACK";
  default: return "??? TRACK";
  }
}

const char* kbdmode(int mode)
{
  switch(mode)
  {
  case 2: return "PROGRAM";
  case 3: return "EDIT PROGRAM";
  case 4: return "GLOBAL";
  case 6: return "SONG";
  case 8: return "DISK";
  case 9: return "EDIT STYLE";
  case 10: return "ARRANGEMENT";
  case 11: return "BACKING SEQUENCE";
  default:
    static char mtxt[4];
    sprintf(mtxt, "%d", mode);
    return (const char*)mtxt;
  }
}