// midi2to0 v1.0 written by Gnter Nagler 1995 (gnagler@ihm.tu-graz.ac.at)
#include "midiio.hpp"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef __MSDOS__
#define WRITE_BINARY  "wb"
#define READ_BINARY   "rb"
#else
#define WRITE_BINARY  "w"
#define READ_BINARY   "r"
#endif

char* input = 0;
char output[256];
char* outprefix = "out";

MidiWrite* write = 0;
int quiet = 0;

class Midi2to0 : public MidiRead
{
public:
  Midi2to0(char* name);

  virtual void head(unsigned version, unsigned tracks, unsigned unitsperbeat);
  virtual void track(int trackno, long length, int channel);
  virtual void endtrack(int trackno);

  virtual void text(int what, int len, char* whattext, unsigned char* txt);
  virtual void meta(int what, int len, unsigned char* data);
  virtual void end();
  virtual void tact(int nom, int denom, int v1, int v2);
  virtual void tempo(unsigned long ticks);
  virtual void program(int prg, int channel);
  virtual void control(int channel, int what, int value);
  virtual void highbank(int channel, int val);
  virtual void wheel(int channel, int val);
  virtual void breath(int channel, int val);
  virtual void foot(int channel, int val);
  virtual void portamentotime(int channel, int val);
  virtual void data(int channel, int val);
  virtual void volume(int channel, int val);
  virtual void balance(int channel, int val);
  virtual void expression(int channel, int val);
  virtual void lowbank(int channel, int val);
  virtual void hold(int channel, int val);
  virtual void reverb(int channel, int val);
  virtual void chorus(int channel, int val);
  virtual void datainc(int channel, int val);
  virtual void datadec(int channel, int val);
  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void time(unsigned long ticks);
  virtual void pitchbend(int channel, int val);
  virtual void polyaftertouch(int channel, int note, int val);
  virtual void aftertouch(int channel, int val);
  virtual void songpos(unsigned pos);
  virtual void songselect(unsigned char song);
  virtual void tunerequest();
  virtual void timingclock();
  virtual void start();
  virtual void cont();
  virtual void stop();
  virtual void activesense();
  virtual void sysex(int syslen, unsigned char* sysdata);
  virtual void error(const char* msg);

  virtual void percent(int perc);
};

Midi2to0::Midi2to0(char* name) : MidiRead(name)
{
}

void Midi2to0::head(unsigned version, unsigned tracks, unsigned unitsperbeat)
{
  if (version != 2)
  {
    fprintf(stderr, "This file is not midi format 2. Its format %d.\n", version);
    exit(1);
  }
}

void Midi2to0::track(int trackno, long length, int channel)
{
  sprintf(output, "%s%d.mid", outprefix, trackno);
  write = new MidiWrite(output);
  if (!write)
  {
    fprintf(stderr, "out of memory\n");
    exit(1);
  }
  if (!write->getf())
  {
    perror(output);
    exit(1);
  }
  fprintf(stderr, "writing %s\n", output);
  write->head(0, 1, unitsperbeat_);
  write->track();
  write->tempo(500000l); // initialize to 120 beats/min
}

void Midi2to0::endtrack(int trackno)
{
  write->endtrack();

  delete write; write = 0;
}

void Midi2to0::text(int what, int len, char* whattext, unsigned char* txt)
{
  write->text(what, len, txt);
}

void Midi2to0::meta(int what, int len, unsigned char* data)
{
  write->meta(what, len, data);
}

void Midi2to0::end()
{
}

void Midi2to0::tact(int nom, int denom, int v1, int v2)
{
  write->tact(nom, denom, v1, v2);
}

void Midi2to0::tempo(unsigned long ticks)
{
  write->tempo(ticks);
}

void Midi2to0::time(unsigned long ticks)
{
  write->time(ticks);
}

void Midi2to0::program(int channel, int prg)
{
  write->program(channel, prg);
}

void Midi2to0::control(int channel, int what, int value)
{
  write->control(channel, what, value);
}

void Midi2to0::balance(int channel, int val)
{
  write->balance(channel, val);
}

void Midi2to0::highbank(int channel, int val)
{
  write->highbank(channel, val);
}

void Midi2to0::wheel(int channel, int val)
{
  write->wheel(channel, val);
}

void Midi2to0::breath(int channel, int val)
{
  write->breath(channel, val);
}

void Midi2to0::foot(int channel, int val)
{
  write->foot(channel, val);
}

void Midi2to0::portamentotime(int channel, int val)
{
  write->portamentotime(channel, val);
}

void Midi2to0::data(int channel, int val)
{
  write->data(channel, val);
}

void Midi2to0::volume(int channel, int val)
{
  write->volume(channel, val);
}

void Midi2to0::expression(int channel, int val)
{
  write->expression(channel, val);
}

void Midi2to0::lowbank(int channel, int val)
{
  write->lowbank(channel, val);
}

void Midi2to0::hold(int channel, int val)
{
  write->hold(channel, val);
}

void Midi2to0::reverb(int channel, int val)
{
  write->reverb(channel, val);
}

void Midi2to0::chorus(int channel, int val)
{
  write->chorus(channel, val);
}

void Midi2to0::datainc(int channel, int val)
{
  write->datainc(channel, val);
}

void Midi2to0::datadec(int channel, int val)
{
  write->datadec(channel, val);
}

void Midi2to0::noteon(int channel, int note, int vel)
{
  write->noteon(channel, note, vel);
}

void Midi2to0::noteoff(int channel, int note, int vel)
{
  write->noteoff(channel, note, vel);
}

void Midi2to0::pitchbend(int channel, int val)
{
  write->pitchbend(channel, val);
}

void Midi2to0::polyaftertouch(int channel, int note, int val)
{
  write->polyaftertouch(channel, note, val);
}

void Midi2to0::aftertouch(int channel, int val)
{
  write->aftertouch(channel, val);
}

void Midi2to0::songpos(unsigned pos)
{
  write->songpos(pos);
}

void Midi2to0::songselect(unsigned char song)
{
  write->songselect(song);
}

void Midi2to0::tunerequest()
{
  write->tunerequest();
}

void Midi2to0::timingclock()
{
  write->timingclock();
}

void Midi2to0::start()
{
  write->start();
}

void Midi2to0::cont()
{
  write->cont();
}

void Midi2to0::stop()
{
  write->stop();
}

void Midi2to0::activesense()
{
  write->activesense();
}

void Midi2to0::sysex(int syslen, unsigned char* sysdata)
{
  write->sysex(syslen, sysdata);
}

void Midi2to0::percent(int perc)
{
  if (!quiet)
    fprintf(stderr, "%-3d%%\r", perc);
}

void Midi2to0::error(const char* msg)
{
  printf("// error: %s\n", msg);
  exit(1);
}

void usage()
{
  fprintf(stderr, "midi2to0 splits midi2 sequences into some midi0 sequences\n");
  fprintf(stderr, "usage: midi2to0 [-q] [-prefix out] midi2fmt.mid\n");
  fprintf(stderr, "-q\tbe quiet\n");
  fprintf(stderr, "-prefix out\tuses out as prefix for output filenames out*.mid\n");
  exit(1);
}

int main(int argc, char**argv)
{
  argc--; argv++;
  while (argc > 0 && **argv == '-')
  {
    if (argc > 0 && strcmp(*argv, "-q") == 0)
    {
      quiet = 1;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strcmp(*argv, "-prefix") == 0)
    {
      argc--; argv++;
      if (argc > 0)
      {
	outprefix = *argv++; argc--;
      }
      else
	fprintf(stderr, "option -prefix needs a parameter\n");
      continue;
    }
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  if (argc <= 0)
    usage();

  input = argv[0];
  Midi2to0 midi(input);
  if (!midi.getf())
  {
    perror(input);
    return 1;
  }
  argc--; argv++;
  if (!midi.run())
    fprintf(stderr, "%s: midi read error at %04lX\n", input, midi.getpos());

  delete write;
  return 0;
}
