static char* version = "midi2txt v1.4 by Gnter Nagler (" __DATE__ ")";

#include "midiio.hpp"
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#ifdef __MSDOS__
#include <dos.h>
#else
#endif

// values for info_
#define ONLY_ERRORS -2
#define ONLY_LYRICS -1
#define ONLY_CHUNK   0
#define ONLY_SHORT   1
#define ONLY_INFO    2
#define ALL_EVENTS   3

const char* versioninfo(int version)
{
  switch(version)
  {
  case 0: return "single multichanneltrack";
  case 1: return "several tracks with seperated channels to play all at once";
  case 2: return "several multichanneltracks to play one by one";
  default: return "unknown version";
  }
}

class MidiPrint : public MidiRead
{
public:
  MidiPrint(char* name);

  virtual void head(unsigned version, unsigned tracks, unsigned unitperbeat);
  virtual void track(int trackno, long length, int channel);
  virtual void endtrack(int trackno);

  virtual void seqnumber(unsigned int seqno);
  virtual void text(int what, int len, char* whattext, unsigned char* txt);
  virtual void meta(int what, int len, unsigned char* data);
  virtual void end();
  virtual void prefixchannel(unsigned char channel);
  virtual void prefixport(unsigned char port);
  virtual void smpteofs(int hour, int min, int sec, int frame, int fracframe);
  virtual void tact(int nom, int denom, int unitsperbeat, int notes32perbeat);
  virtual void tempo(unsigned long ticks);
  virtual void key(int signature, int isminor);

  virtual void program(int prg, int channel);
  virtual void control(int channel, int what, int value);
    virtual void highbank(int channel, int val);
    virtual void wheel(int channel, int val);
    virtual void breath(int channel, int val);
    virtual void foot(int channel, int val);
    virtual void portamentotime(int channel, int val);
    virtual void data(int channel, int val);
    virtual void volume(int channel, int val);
    virtual void balance(int channel, int val);
    virtual void expression(int channel, int val);
    virtual void lowbank(int channel, int val);
    virtual void hold(int channel, int val);
    virtual void reverb(int channel, int val);
    virtual void chorus(int channel, int val);
    virtual void datainc(int channel, int val);
    virtual void datadec(int channel, int val);
    virtual void lowrpn(int channel, int val);
    virtual void highrpn(int channel, int val);
    virtual void pitchbendrange(int channel, int val);
  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void time(unsigned long ticks);
  virtual void pitchbend(int channel, int val);
  virtual void polyaftertouch(int channel, int note, int val);
  virtual void aftertouch(int channel, int val);
  virtual void songpos(unsigned pos);
  virtual void songselect(unsigned char song);
  virtual void tunerequest();
  virtual void timingclock();
  virtual void start();
  virtual void cont();
  virtual void stop();
  virtual void activesense();
  virtual void sysex(int syslen, unsigned char* sysdata);
  virtual void gmreset();
  virtual void gsreset();
  virtual void gsexit();
  virtual void xgreset();
  virtual void endmidi();
  virtual void error(const char* msg);

  virtual void percent(int perc);

  void printchannel(int channel);

  int info_;
  int marktact_;
  long errorcnt_;

private:
  int indent_;
  unsigned tactunits_;

  int nl_;
  int endmark_;  // each track should contain FF 2F 00 as last command

  void indent();
};

MidiPrint::MidiPrint(char* name) : MidiRead(name, NULL, MIDI_PRELOAD)
{
  indent_ = 0;
  info_ = ALL_EVENTS;
  marktact_ = 0;
  tactunits_ = 0;
  nl_ = 1;
  errorcnt_ = 0;
}

void MidiPrint::indent()
{
 for (; indent_ > 0; indent_--)
    putchar(' ');
}

void MidiPrint::head(unsigned version, unsigned tracks, unsigned unitperbeat)
{
  if (midiname_)
    printf("// %s\n", midiname_);
  if (getcurpos() != 14)
    printf(" // Warning: %ld bytes garbage at beginning of file\n", getcurpos() - 14);
  if (tracks_ == 0 && info_ == ONLY_ERRORS)
    error("no track found");
  if (info_ == ONLY_SHORT || info_ == ONLY_LYRICS || info_ == ONLY_ERRORS)
    return;
  printf("mthd\n");
  printf("  version %d // %s\n", version, versioninfo(version));
  printf("  // %d track%s\n", tracks, tracks != 1 ? "s" : "");
  printf("  unit %d // is 1/4\n", unitperbeat);
  printf("end mthd\n");
}

void MidiPrint::track(int trackno, long length, int channel)
{
  endmark_ = 0;
  if (info_ == ONLY_SHORT || info_ == ONLY_LYRICS || info_ == ONLY_ERRORS)
    return;
  printf("\nmtrk");
  if (channel >= 0)
    printf("($%X)\n", channel+1);
  else if (channel == MULTICHANNEL)
    printf(" // multichannel track\n");
  else
    printf("\n");
}

void MidiPrint::endtrack(int trackno)
{
  if (!endmark_)
    printf(" // Warning: end of track meta event missing\n");

  if (info_ == ONLY_SHORT || info_ == ONLY_ERRORS)
    return;
  if (info_ == ONLY_LYRICS)
  {
    if (!nl_)
    {
      nl_ = 1;
      printf("\n");
    }
    return;
  }
  if (info_ >= ALL_EVENTS)
    printf("\n");
  printf("end mtrk\n");
}

void printchar(int c)
{
  if (c < 0)
    return;
  if (c == '\\')
    printf("\\\\");
  else if (c == '"')
    printf("\\\"");
  else if (c != 0 && (isprint(c) || strchr("", c) != 0))
    putchar(c);
  else
    printf("\\x%02x", c);
}

void MidiPrint::seqnumber(unsigned int seqno)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printf("seqnumber %u\n", seqno);
}

void MidiPrint::text(int what, int len, char* whattext, unsigned char* txt)
{
  if (info_ == ONLY_ERRORS)
    return;
  if (info_ == ONLY_LYRICS)
  {
    if (!txt)
      return;
    if (what == meta_lyric || what == meta_text)
    {
    char* bs;

      while ((bs = (char*)memchr(txt, '\\', len)) != 0)
	*bs = '\n';
      while ((bs = (char*)memchr(txt, '/', len)) != 0)
	*bs = '\n';
      if (*txt == '@')
      {
	txt++; len--;
	switch(*txt)
	{
	case 'K': printf("\nKind: "); break;
	case 'V': printf("\nVersion: "); break;
	case 'I': printf("\nInformation: "); break;
	case 'T': printf("\nTitle: "); break;
	case 'L': printf("\nLanguage: "); break;
	default:  printf("\n@%c: ", *txt); break;
	}
	if (*txt)
	{
	  txt++;
	  len--;
	}
      }
      if (txt[len-1] == '\r')
	txt[len-1] = '\n';
      nl_ = txt[len-1] == '\n';
      printf("%.*s", len, txt);
    }
    return;
  }
  indent();
  if (whattext)
    printf("%s ", whattext);
  else
    printf("metaevent %d(%d) ", what, len);

  putchar('"');
  if (txt)
  {
    while (len-- > 0)
      printchar(*txt++);
  }
  putchar('"');
  if (!whattext)
    printf(" end metaevent");
  printf("\n");
}

void MidiPrint::meta(int what, int len, unsigned char* data)
{
int i;

  if (info_ < ALL_EVENTS)
    return;
  indent();
  printf("metaevent %d(%d)", what, len);
  i = 0;
  while (len-- > 0)
  {
    indent_ = 4;
    printf(" $%02X", *data++);
    if ( (i % 16) == 15)
    {
      printf("\n");
      indent();
    }
    i++;
  }
  printf(" end metaevent\n");
}

void MidiPrint::end()
{
  endmark_ = 1;
}

void MidiPrint::prefixchannel(unsigned char channel)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printf("prefixchannel %d\n", channel+1);
}

void MidiPrint::prefixport(unsigned char port)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printf("prefixport %d\n", port);
}


void MidiPrint::smpteofs(int hour, int min, int sec, int frame, int fracframe)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printf("smpteofs %02d:%02d:%02d:%d:%d\n",
    hour, min, sec, frame, fracframe);
}

void MidiPrint::tact(int nom, int denom, int unitsperbeat, int notes32perbeat)
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  printf("tact %d / %d %d %d\n", nom, denom, unitsperbeat, notes32perbeat);
  if (!marktact_)
    return;
  if (tactunits_)
  {
    tactunits_ = 0;
    marktact_ = 0;
  }
  else
    tactunits_ = unitsperbeat_ * 4 * nom / denom;
}

void MidiPrint::tempo(unsigned long ticks)
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  if (ticks != 0)
  {
      printf("beats %2.5f\n", (float)(60000000.0 / (float)ticks));
      return;
  }
  else
    printf("tempo %ld\n", ticks);
}

void MidiPrint::key(int signature, int isminor)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printf("key \"");
  if (signature != 0)
  {
    printf("%d%c ", (signature < 0) ? -signature : signature,
	(signature < 0) ? 'b' : '#');
  }
  else
    printf("C");
  printf("%s\"\n",  isminor ? "min" : "maj");
}

void MidiPrint::printchannel(int channel)
{
  if (info_ < ALL_EVENTS)
    return;
  if (getchannel() != channel)
  {
    indent();
    if (channel >= 10)
      printf("[$%X]", channel+1);
    else if (channel >= 0)
      printf("[%d]", channel+1);
    setchannel(channel);
  }
}

void MidiPrint::time(unsigned long ticks)
{
unsigned long n;

//  fprintf(stderr, "%lu:%02d\r", getcurmillisec()/ 60000L, int((getcurmillisec() / 1000) % 60));
  indent_ = 2;
  while (ticks > 0)
  {
    if (!marktact_ || !tactunits_)
      n = ticks;
    else
    {
      if (curunit_ % tactunits_ == 0)
	n = tactunits_;
      else
	n = tactunits_ - curunit_ % tactunits_;
      if (n > ticks)
	n = ticks;
    }
    if (info_ < ALL_EVENTS)
      return;
    if (n > 0)
    {
      indent();
      curunit_ += n;
      if ((n % unitsperbeat_) == 0)
	printf("%ld/4", n / unitsperbeat_);
      else if (unitsperbeat_ % 2 == 0 && (n % (unitsperbeat_ / 2)) == 0)
	printf("%ld/8", n / (unitsperbeat_/2));
      else if (unitsperbeat_ % 4 == 0 && (n % (unitsperbeat_ / 4)) == 0)
	printf("%ld/16", n / (unitsperbeat_/4));
      else
	printf("%ld", n);
      if (!marktact_ || !tactunits_ || (curunit_ % tactunits_) != 0)
	putchar(';');
      else
      {
	printf(". // tact %lu\n", curunit_ / tactunits_);
	indent_ = 2;
      }
      ticks -= n;
    }
  }
}

void MidiPrint::program(int channel, int prg)
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  printchannel(channel);
  printf("program %s\n", progname(prg, channel));
}

void MidiPrint::control(int channel, int what, int value)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("control $%02X %d\n", what, value);
}

void MidiPrint::balance(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("balance ");
  if (val == 0)
    printf(" left\n");
  else if (val == 127)
    printf(" right\n");
  else
    printf(" %d\n", val);
}

void MidiPrint::highbank(int channel, int val)
{
  if (info_ < ONLY_INFO)
    return;
  indent();
  printchannel(channel);
  printf("hbank $%02X\n", val);
}

void MidiPrint::wheel(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("wheel %d\n", val);
}

void MidiPrint::breath(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("breath %d\n", val);
}

void MidiPrint::foot(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("foot %d\n", val);
}

void MidiPrint::portamentotime(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("portamentotime %d\n", val);
}

void MidiPrint::data(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("data %d\n", val);
}

void MidiPrint::volume(int channel, int val)
{
  if (info_ < ONLY_INFO)
    return;
  indent();
  printchannel(channel);
  printf("volume %d\n", val);
}

void MidiPrint::expression(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("expression %d\n", val);
}

void MidiPrint::lowbank(int channel, int val)
{
  if (info_ < ONLY_INFO)
    return;
  indent();
  printchannel(channel);
  printf("lbank $%02X\n", val);
}

void MidiPrint::hold(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("hold ");
  if (val == 0)
    printf(" off\n");
  else if (val == 0x7f)
    printf(" on\n");
  else
    printf(" %d\n", val);
}

void MidiPrint::reverb(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("reverb %d\n", val);
}

void MidiPrint::chorus(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("chorus %d\n", val);
}

void MidiPrint::datainc(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("datainc %d\n", val);
}

void MidiPrint::datadec(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("datadec %d\n", val);
}

void MidiPrint::lowrpn(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("lowrpn %d\n", val);
}

void MidiPrint::highrpn(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("highrpn %d\n", val);
}

void MidiPrint::pitchbendrange(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("pitchbendrange %d\n", val);
}

void MidiPrint::noteon(int channel, int note, int vel)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("+%s $%02X;\n", notename(note), vel);
}

void MidiPrint::noteoff(int channel, int note, int vel)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("-%s $%02X;\n", notename(note), vel);
}

void MidiPrint::pitchbend(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("pitchbend %d\n", val);
}

void MidiPrint::polyaftertouch(int channel, int note, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("polyaftertouch %s %d\n", notename(note), val);
}

void MidiPrint::aftertouch(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  printf("aftertouch %d\n", val);
}

void MidiPrint::songpos(unsigned pos)
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  printf("songpos %d\n", pos);
}

void MidiPrint::songselect(unsigned char song)
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  printf("songselect %d\n", song);
}

void MidiPrint::tunerequest()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printf("tunerequest\n");
}

void MidiPrint::timingclock()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printf("timingclock\n");
}

void MidiPrint::start()
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  printf("start\n");
}

void MidiPrint::cont()
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  printf("continue\n");
}

void MidiPrint::stop()
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  printf("stop\n");
}

void MidiPrint::activesense()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printf("activesensing\n");
}

void MidiPrint::sysex(int syslen, unsigned char* sysdata)
{
int i = 0;

  if (info_ < ALL_EVENTS)
    return;
  indent();
  syslen--; // last byte is end of sysex (F7)
  printf("sysevent(%d) ", syslen);
  while (syslen-- > 0)
  {
    indent_ = 4;
    printf("$%02X ", *sysdata++);
    if ( (i % 16) == 15)
    {
      printf("\n");
      indent();
    }
    i++;
  }
  printf("end sysevent\n");
}

void MidiPrint::gmreset()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printf("GMReset\n");
}

void MidiPrint::gsreset()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printf("GSReset\n");
}

void MidiPrint::xgreset()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printf("XGReset\n");
}

void MidiPrint::gsexit()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printf("GSExit\n");
}

void MidiPrint::endmidi()
{
  if (info_ < ALL_EVENTS && info_ != ONLY_ERRORS)
    return;
  if (getpos() < filesize_)
  {
  long len = filesize_ - getpos();

    printf(" // Warning: %ld bytes garbage at end of file\n", len);
    if (info_ != ONLY_ERRORS)
    {
      while (len > 0)
      {
      int n = 16;

	if (len < n)
	  n = (int)len;
	unsigned char* c = get(n);
	if (!c)
	  return;
	printf("//");
	for (int i = 0; i < n; i++)
	  printf(" %02X", c[i]);
	printf("\n");
	len -= n;
      }
    }
  }
}

void MidiPrint::percent(int perc)
{
  if (info_ >= ONLY_CHUNK)
  fprintf(stderr, "%-3d%%\r", perc);
}

void MidiPrint::error(const char* msg)
{
  printf("// error: %s\n", msg);
  errorcnt_++;
}

void usage()
{
  fprintf(stderr, "usage: midi2txt [-error][-chunk][-short][-info][-lyric][-tact] file.mid [> file.txt]\n");
  exit(1);
}

int main(int argc, char**argv)
{
int info = ALL_EVENTS;
int tact = 0;
char* midiname = 0;

  argc--; argv++;

  while (argc > 0 && **argv == '-')
  {
    if (strncmp(*argv, "-version", 2) == 0)
    {
      fprintf(stderr, "%s\n", version);
      argc--; argv++;
      if (argc == 0)
	return 0;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-error", 2) == 0)
    {
      info = ONLY_ERRORS; // check if errors in midi
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-chunk", 2) == 0)
    {
      info = ONLY_CHUNK;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-short", 2) == 0)
    {
      info = ONLY_SHORT;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-info", 2) == 0)
    {
      info = ONLY_INFO;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-lyric", 2) == 0)
    {
      info = ONLY_LYRICS;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-tact", 2) == 0)
    {
      tact = 1;
      argc--; argv++;
      continue;
    }
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  if (argc == 0)
    usage();

  midiname = *argv;
  MidiPrint midi(midiname);
  if (!midi.getf())
  {
    perror(midiname);
    return 1;
  }
  fprintf(stderr, "%s\n", midiname);
  midi.info_ = info;
  midi.marktact_ = tact;
  if (info == ONLY_CHUNK)
    midi.options_ = OPTION_NOEVENTS;
  if (!midi.run())
  {
    printf("%s: midi error at 0x%04lx\n", midiname, midi.getpos());
    return 1;
  }
  return 0;
}

