// mididir written by Gnter Nagler 1996 (gnagler@ihm.tu-graz.ac.at)
#include "midiio.hpp"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifdef __MSDOS__
#include <dir.h>
#include <dos.h>
#include <conio.h>
#ifdef GNDIR
#include "gndir.h"
#endif
#endif

static char* version = "mididir v1.4 by Gnter Nagler (" __DATE__ ")";

#ifdef __MSDOS__
#define WRITE_BINARY  "wb"
#define READ_BINARY   "rb"
#else
#define WRITE_BINARY  "w"
#define READ_BINARY   "r"
#endif

char* addtitle = 0;
char* searchtitle = 0;
char* input = 0;
char* output = "temp0000.mid";
int database = 0;
long count = 0;
long notmidi = 0;

#define TIME_LIMIT   1000  // useful information are before this midi unit

int match(char* text, int textlen, char* r, int rlen);
int matchsub(char* text, int textlen, char* r, int rlen);
int matchsub(char* text, char* r);

class MidiTitle : public MidiCopy // add midi title
{
public:
  MidiTitle(char* name);

  virtual void track(int trackno, long length, int channel);

  virtual void error(const char* msg);
  virtual void warning(const char* msg);

  virtual void percent(int perc);
};

MidiTitle::MidiTitle(char* name) : MidiCopy(name)
{
}

void MidiTitle::track(int trackno, long length, int channel)
{
  MidiCopy::track(trackno, length, channel);
  if (trackno == 1 && dest_ && addtitle && *addtitle)
    dest_->text(meta_trackname, -1, addtitle);
}

void MidiTitle::percent(int perc)
{
    fprintf(stderr, "%-3d%%\r", perc);
}

void MidiTitle::error(const char* msg)
{
  printf("Error: %s\n", msg);
  exit(1);
}

void MidiTitle::warning(const char* msg)
{
  printf("Warning: %s\n", msg);
}

class MidiScanTitle : public MidiRead
{
public:
  MidiScanTitle(char* name);

  virtual void text(int what, int len, char* whattext, unsigned char* txt);

  virtual void xgreset();
  virtual void gmreset();
  virtual void gsreset();

  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);

  virtual void endmidi();

  virtual void error(const char* msg);
  virtual void warning(const char* msg);

  virtual void percent(int perc);

  char xg_;
  char gs_;
  char gm_;

  char title_[256];
  char err_[256];

protected:
  char text_[256];
  char karaoke_[256];

};

MidiScanTitle::MidiScanTitle(char* name) : MidiRead(name)
{
  options_ = OPTION_NOCONTROLS;
  strcpy(err_, "");
  strcpy(title_, "");
  strcpy(text_, "");
  strcpy(karaoke_, "");
  xg_ = gs_ = gm_ = 0;
}

int unreadable(char* s)
{
int invalid = 0;

  while (*s)
  {
    if ((*s < ' ' && !isspace(*s)) || *s >= 0x7F)
    {
      invalid++;
    }
    s++;
  }
  return invalid > 2;
}

int badtitle(char* s)
{
  if (!s)
    return 1;

  while (*s && isspace(*s))
    s++;
  if (*s == 0)
    return 1;

  if (unreadable(s))
    return 1;
  if (matchsub(s, "WinJammer Demo"))
    return 1;
  if (matchsub(s, "Noteworthy Composer"))
    return 1;
  if (matchsub(s, "MidiScan"))
    return 1;
  if (matchsub(s, "New track"))
    return 1;
  if (matchsub(s, "Conduct"))
    return 1;
  if (matchsub(s, "Seq-"))
    return 1;
  if (matchsub(s, "untitled"))
    return 1;
  if (matchsub(s, "call*BBS"))
    return 1;
  if (matchsub(s, "distribut"))
    return 1;
  if (matchsub(s, "??@*.")) // mail address
    return 1;
  if (matchsub(s, "http://")) // WWW address
    return 1;
  if (matchsub(s, "ftp://")) // ftp address
    return 1;
  return 0;
}

void MidiScanTitle::text(int what, int len, char* whattext, unsigned char* txt)
{
static char t[256];

  if (trackno_ == 1 && title_[0] == '\0')
  {
    if (what == meta_trackname)
    {
      if (len > sizeof(title_))
        len = sizeof(title_)-1;
      if (len > 0)
      {
	  strncpy(t, txt, len);
	  t[len] = '\0';
	  if (!badtitle(t))
	    strcpy(title_, t);
      }
    }
  }

  if (what == meta_text && karaoke_[0] == 0 && len >= 2 && len > 2 && txt[0] == '@' && txt[1] == 'T')
  {
    txt += 2;
    len -=2;

    if (len > sizeof(karaoke_))
      len = sizeof(karaoke_)-1;
    if (len > 0)
    {
	strncpy(t, txt, len);
	t[len] = '\0';
	if (!badtitle(t))
	  strcpy(karaoke_, t);
    }
  }

  if (what == meta_text && trackno_ == 1 && title_[0] == 0)
  {
    if (len > sizeof(text_))
      len = sizeof(text_)-1;
    if (len > 0)
    {
	strncpy(t, txt, len);
	t[len] = '\0';
	if (!badtitle(t))
	  strcpy(text_, t);
    }
  }

  if (text_[0] && karaoke_[0])
     options_ |= OPTION_NOEVENTS;
}

void MidiScanTitle::noteon(int , int , int )
{
  if (getcurunit() > TIME_LIMIT)
    skiptrack_ = 1;
}

void MidiScanTitle::noteoff(int channel, int note, int vel)
{
  if (getcurunit() > TIME_LIMIT)
    skiptrack_ = 1;
}

void MidiScanTitle::xgreset()
{
  xg_ = 1;
}

void MidiScanTitle::gmreset()
{
  gm_ = 1;
}

void MidiScanTitle::gsreset()
{
  gs_ = 1;
}


void MidiScanTitle::endmidi()
{
  if (karaoke_[0])
    strcpy(title_, karaoke_);
  else if (title_[0] == 0 && text_[0] != 0)
    strcpy(title_, text_);
}

void MidiScanTitle::percent(int perc)
{
}

void MidiScanTitle::error(const char* msg)
{
  strcpy(err_, msg);
}

void MidiScanTitle::warning(const char* msg)
{
  strcpy(err_, msg);
}

int match(char* text, int textlen, char* r, int rlen)
{
  while (rlen > 0)
  {
    if (*r == '?')
    {
      if (textlen == 0)
        return 0;
      r++; rlen--;
      text++; textlen--;
    }
    else if (*r == '*')
    {
      while (rlen > 0 && *r == '*')
      {
        r++; rlen--;
      }
      for (int i = 0; i <= textlen; i++)
      if (match(text+i, textlen-i, r,rlen))
        return 1;
      return 0;
    }
    else
    {
      if (textlen == 0)
        return 0;
      if (toupper(*text) != toupper(*r))
        return 0;
      r++; rlen--;
      text++; textlen--;
    }
  }
  return textlen == 0;
}

int matchsub(char* text, char* r)
{
  return matchsub(text, strlen(text), r, strlen(r));
}

int matchsub(char* text, int textlen, char* r, int rlen)
{
char p[256];

  p[0] = '*';
  strncpy(p+1, r, rlen);
  p[rlen+1] = '*';
  p[rlen+2] = 0;
  return match(text, textlen, p, rlen+2);
}

int match(char* text, char* expr)
{
char* word;

  if (!text || !expr)
    return 0;

  int textlen = strlen(text);
  // extract word from expr

  while (*expr)
  {
    while (*expr && isspace(*expr))
      expr++;
    word = expr;
    while (*expr != 0 && !isspace(*expr))
      expr++;
    if (!matchsub(text, textlen, word, (int)(expr-word)))
	return 0;
  }
  return 1;
}

char* replacestr(char* s, char* oldchars, char newchar)
{
  if (!s || !oldchars)
    return s;

  for (char* p = s; *p; p++)
  if (strchr(oldchars, *p))
    *p = newchar;
  return s;
}

void midifile(char* filename)
{
MidiScanTitle midi(filename);

  if (!midi.getf())
  {
    perror(filename);
    return;
  }
  long filelen;

  fseek(midi.getf(), 0, SEEK_END);
  filelen = ftell(midi.getf());

  if (filelen < 14 || !midi.runhead())
  {
    notmidi++;
    return; // not standard midi
  }

  midi.run();

  if (searchtitle && *searchtitle)
  {
    if (!match(midi.title_, searchtitle))
      return;
  }

  const char* name;

#if defined( __MSDOS__) && defined(GNDIR)
  name = getlongfilename(filename);
  if (!name)
    name = filename;
#else
  name = filename;
#endif

  replacestr(midi.title_, "\n\r\t", ' ');
  if (database)
  {
    if (count == 0)
      printf("Filename;Title/Artist;Filesize;Format;Tracks;Error;Specialmode\n");

    replacestr(midi.title_, ";", ',');

    printf("%s;%s;%ld;%d;%d;%s;",
      name,
      midi.title_,
      filelen,
      midi.version_,
      midi.tracks_,
      midi.err_);
     if (midi.gm_)
       printf("GM");
     if (midi.xg_)
       printf(" XG");
     if (midi.gs_)
       printf(" GS");
     printf("\n");
  }
  else
  {
    printf("%s:\t%s (%ld kb, format %d, %d track%s%s%s",
     name, midi.title_,
     (filelen +1023)/ 1024,
     midi.version_,
     midi.tracks_, midi.tracks_ == 1 ? "" : "s",
     midi.err_[0] ? ", " : "", midi.err_);

     if (midi.gm_)
       printf(", GM");
     if (midi.xg_)
       printf(", XG");
     if (midi.gs_)
       printf(", GS");
     printf(")\n");
  }
  count++;
}

char* getname(char* pathname)
{
    char* name = strrchr(pathname, '\\');
    if(name)
      return name+1;
    else
    {
      name = strrchr(pathname, ':');
      if (name)
        return name+1;
    }
    return pathname;
}

void mididir(char* filemask)
{
  if (!filemask || !*filemask)
    return;
#ifdef __MSDOS__
  if (kbhit() && getch() == 27) // escape under DOS only
    exit(1);
  struct ffblk ff;
  char pathname[128];
  int done;

  strcpy(pathname, filemask);
  char* name = getname(pathname);
  char* mask = getname(filemask);

  int isdir = 0;

  if (pathname[strlen(pathname)-1] == ':')
     isdir = 1;
  else if (pathname[strlen(pathname)-1] == '\\')
     isdir = 1;
  else
     isdir = (findfirst(pathname, &ff, FA_DIREC) == 0 && (ff.ff_attrib & FA_DIREC) != 0 && ff.ff_name[0] != '.');
  if (isdir)
  {
    char lastc = pathname[strlen(pathname)-1];
    if (lastc != '\\' && lastc != ':')
      strcat(pathname, "\\");
    strcat(pathname, "*.*");
    name = getname(pathname);
    mask = "*.*";
  }

  done = findfirst(pathname, &ff, 0);
  while (!done)
  {
    if (kbhit() && getch() == 27) // escape under DOS only
      exit(1);
    strcpy(name, ff.ff_name);
    if ((ff.ff_attrib & (FA_DIREC+FA_LABEL)) == 0)
    {
      midifile(pathname);
    }
    done = findnext(&ff);
  }

  // subdirs
  strcpy(name, "*.*");
  done = findfirst(pathname, &ff, FA_DIREC);
  while (!done)
  {
    strcpy(name, ff.ff_name);
    if ((ff.ff_attrib & FA_DIREC) != 0 && ff.ff_name[0] != '.')
    {
      // subdirs;
      strcat(pathname, "\\");
      strcat(pathname, mask);
      mididir(pathname);
    }
    done = findnext(&ff);
  }
#else
  midifile(filemask);
#endif
}

void copyfile(char* input, char* output)
{
FILE* i = fopen(input, READ_BINARY);

  if (!i)
  {
    perror(input);
    return;
  }
FILE* o = fopen(output, WRITE_BINARY);
  if (!o)
  {
    fclose(i);
    perror(output);
    return;
  }
  static char buf[4096];
  int n;
  long pos = 0;

  while (fread(buf, sizeof(buf), 1, i) == 1)
  {
    fwrite(buf, sizeof(buf), 1, o);
    pos += sizeof(buf);
  }
  fseek(i, pos, SEEK_SET);
  fseek(o, pos, SEEK_SET);
  while ((n = fread(buf, 1, sizeof(buf), i)) > 0)
    fwrite(buf, n, 1, o);

  fclose(i);
  fclose(o);
}

void usage()
{
  printf("mididir shows midi directory listing\n");
  printf("usage: mididir [-h] [-t \"title\"] [-s \"words\"] [-database] files ...\n");
  printf("-h\t\tusage help\n");
  printf("-version\tget program version\n");
  printf("-t \"title\"\tadds a title to the midi file\n");
  printf("-s \"words\"\tfind midi files having these words in their title\n");
  printf("-database\tproduce database importable format\n");
  exit(1);
}

int main(int argc, char**argv)
{
  argc--; argv++;
  while (argc > 0 && **argv == '-')
  {
    if (strncmp(*argv, "-help", 2) == 0)
    {
      usage();
      argc--; argv++; continue;
    }
    if (strncmp(*argv, "-version", 2) == 0)
    {
      fprintf(stderr, "%s\n", version);
      argc--; argv++;
      if (argc == 0)
	return 0;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-search", 2) == 0)
    {
      argc--; argv++;
      if (argc > 0)
      {
        searchtitle = *argv++; argc--;
      }
      else
        fprintf(stderr, "option -s needs parameter\n");
      continue;
    }
    if (strncmp(*argv, "-database", 2) == 0)
    {
      database = 1;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-title", 2) == 0)
    {
      argc--; argv++;
      if (argc > 0)
      {
        addtitle = *argv++; argc--;
      }
      else
        fprintf(stderr, "option -t needs a title as parameter\n");
      continue;
    }
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  if (addtitle)
  {
    int writeok = 1;

    if (argc != 1)
    {
      if (argc > 1)
        fprintf(stderr, "mididir: modifying title of multiple files not allowed!\n");
      usage();
      return 1;
    }
    input = *argv++; argc--;
    MidiTitle *midi = new MidiTitle(input);
    if (!midi->getf())
    {
      perror(input);
      return 1;
    }
    MidiWrite* write = new MidiWrite(output);
    if (!write)
    {
      fprintf(stderr, "out of memory\n");
      return 1;
    }
    if (!write->getf())
    {
      perror(output);
      return 1;
    }
    argc--; argv++;

    midi->setoutput(write);

    midi->options_ = OPTION_NOSYSEVENTS+OPTION_NOCONTROLS;
    if (!midi->run())
    {
      writeok = 0;
      fprintf(stderr, "%s: midi read error at %04lX\n", input, midi->getpos());
    }
    delete write;
    delete midi;
    if (writeok)
    {
      copyfile(output, input);
      unlink(output);
    }
  }
  else
  {    if (argc == 0)
      mididir("*.*");
    else
    {
      for (; argc > 0; argc--)
         mididir(*argv++);
    }
  }
  if (notmidi > 0)
    fprintf(stderr, "Warning: %ld files ignored (they are not standard midi files)\n", notmidi);
  return 0;
}
