// mididmp v1.0 written by Gnter Nagler 1995 (gnagler@ihm.tu-graz.ac.at)
#include "mtrack.hpp"
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#ifdef __MSDOS__
#include <alloc.h>
#endif

class MidiDump;

typedef struct
{
  long time;
  int nom, denom;
  int units;
} Tact;

class Note
{
public:
  Note();

  long since_;  // start time
  long until_;  // end time
  int noteval_;
  int vel_;
};


char* input = 0;

int trackcount = 0;
MidiDump *trackread = 0;

long measure = 0;
long curtacttime = 0, nexttacttime = 0;
int curtactdelay = 0;
Tact* tactchangelist = 0;
int curtactchange = 0;
int tactchangecount = 0;
int unitsperquarter = 0;

#define MAXTRACKS 10
int showtracks = 0;
int option_track[MAXTRACKS];
int option_trackcnt = 0;

int *dumptrack = 0;
int dumpchannel[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

Note::Note()
{
  since_ = until_ = -1;
  noteval_ = vel_ = -1;
}

#define MAXNOTESPERTACT  256

void nexttact();
void dumptact(MidiDump*);

int unitspertact(int nom, int denom, int unitsperquarter)
{
  assert(nom > 0 && denom > 0 && unitsperquarter > 0);
  if (denom == 4)
    return unitsperquarter * nom;
  return (4 * unitsperquarter * nom) / denom;
}

class MidiScan : public MidiRead
{
public:
  MidiScan(char* name);

  virtual void head(unsigned version, unsigned tracks, unsigned unitsperbeat);
  virtual void track(int trackno, long length, int channel);
  virtual void endtrack(int trackno);
  virtual void end();

  virtual void error(const char* msg);
  virtual void tact(int nom, int denom, int v1, int v2);
  virtual void text(int what, int len, char* whattext, unsigned char* txt);

  virtual void program(int channel, int prg);

  char tracktitle_[40];
  int program_;
};

MidiScan::MidiScan(char* name) : MidiRead(name)
{
  options_ = OPTION_NOCONTROLS;
}

void MidiScan::head(unsigned version, unsigned tracks, unsigned unitsperbeat)
{
  if (version != 0 && version != 1)
  {
    fprintf(stderr, "%s is format %d. This program only accepts formats 0 and 1. Try to convert it using midi%dto1!\n", midiname_, version);
    exit(1);
  }
  unitsperquarter = unitsperbeat;

  printf("%d units per quarternote\n", unitsperquarter);
  trackcount = tracks;
  assert(trackcount > 0);
  tact(4, 4, 24, 8); // default tact

  dumptrack = (int*)malloc(trackcount * sizeof(int));
  assert(dumptrack != 0);
  for (int i = 0; i < trackcount; i++)
    dumptrack[i] = 0;

  int dumptracks = 0;
  for (i = 0; i < option_trackcnt; i++)
  {
    dumptrack[ option_track[i] -  1] = 1;
    dumptracks++;
  }
  if (!dumptracks)
  for (i = 0; i < trackcount; i++)
    dumptrack[i] = 1;
}

void MidiScan::end()
{
  // sort tactchangelist by time, simple bubblesort

  while (tactchangecount > 1)
  {
    int changed = 0;

    for (int i = 0; i < tactchangecount-1; i++)
    if (tactchangelist[i].time > tactchangelist[i+1].time)
    {
      changed++;
      // swap i, i+1
      Tact temp;

      temp = tactchangelist[i];
      tactchangelist[i] = tactchangelist[i+1];
      tactchangelist[i+1] = temp;
    }

    if (!changed)
      break;
  }
}

void MidiScan::track(int trackno, long length, int channel)
{
  strcpy(tracktitle_, "");
  program_ = -1;
}

void MidiScan::endtrack(int trackno)
{
}

void MidiScan::program(int channel, int prg)
{
  if (program_ < 0)
    program_ = prg;
}

void MidiScan::text(int what, int len, char* whattext, unsigned char* txt)
{
  if (what == meta_trackname)
  {
    if (txt)
    {
      if (len >= sizeof(tracktitle_))
	len = sizeof(tracktitle_)-1;
      strncpy(tracktitle_, txt, len);
      tracktitle_[len] = '\0';
    }
  }
}

void MidiScan::tact(int nom, int denom, int v1, int v2)
{
  for (int i = 0; i < tactchangecount; i++)
  if (tactchangelist[i].time == curunit_)
  {
    tactchangelist[i].nom = nom;
    tactchangelist[i].denom = denom;
    tactchangelist[i].units = unitspertact(nom, denom, unitsperbeat_);
    return;
  }
  if ((tactchangecount % 16) == 0)
  {
    tactchangelist = (Tact*)realloc(tactchangelist, sizeof(Tact) * (tactchangecount+16));
    assert(tactchangelist != 0);
  }
  tactchangelist[tactchangecount].time = curunit_;
  tactchangelist[tactchangecount].nom = nom;
  tactchangelist[tactchangecount].denom = denom;
  tactchangelist[tactchangecount].units = unitspertact(nom, denom, unitsperbeat_);
  tactchangecount++;
}

void MidiScan::error(const char* msg)
{
  printf("// error: %s\n", msg);
  exit(1);
}

long getnexttacttime()
{
  return curtacttime + curtactdelay;
}

void nexttact()
{
int epsilon = 3; // tact change some units too late or too early

  curtacttime += curtactdelay;
  if (curtactchange < tactchangecount && (curtacttime >= tactchangelist[curtactchange].time-epsilon))
     curtactdelay = tactchangelist[curtactchange++].units;
  nexttacttime = getnexttacttime();
  measure++;
}

class MidiDump : public MidiRead
{
public:
  MidiDump(char* name, FILE* f = 0);

  virtual void track(int trackno, long length, int channel);
  virtual void endtrack(int trackno);

  virtual void program(int channel, int prg);
  virtual void control(int channel, int what, int value);
  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void pitchbend(int channel, int val);
  virtual void polyaftertouch(int channel, int note, int val);
  virtual void aftertouch(int channel, int val);
  virtual void songpos(unsigned pos);
  virtual void songselect(unsigned char song);
  virtual void tunerequest();
  virtual void timingclock();
  virtual void start();
  virtual void cont();
  virtual void stop();
  virtual void activesense();
  virtual void sysex(int syslen, unsigned char* sysdata);
  virtual void meta(int what, int len, unsigned char* data);
  virtual void error(const char* msg);


  Note tactnote_[MAXNOTESPERTACT];
  int notecount_;
};

MidiDump::MidiDump(char* name, FILE* f) : MidiRead(name, f)
{
  options_ = OPTION_NOCONTROLS+OPTION_NOMETAEVENTS+OPTION_NOSYSEVENTS;
  notecount_ = 0;
}

void MidiDump::track(int trackno, long length, int channel)
{
int dumpit = 1;

  if (!dumptrack[trackno-1])
    dumpit = 0;
  else if (channel >= 0 && !dumpchannel[channel])
    dumpit = 0;
  else if (channel < 0)
  {
    for (int i = 0; i < 16; i++)
      if (dumpchannel[i])
	break;
    if (i >= 16)
      dumpit = 0;
  }

  options_ &= ~OPTION_NOEVENTS;
  if (dumpit)
  {
    printf("Track %d:\n", trackno);
    curtactchange = 0;
    curtacttime = 0;
    curtactdelay = tactchangelist[curtactchange++].units;
    nexttacttime = getnexttacttime();
    measure = 1;
    return;
  }
  options_ |= OPTION_NOEVENTS;
}

void MidiDump::endtrack(int trackno)
{
  dumptact(this);

  options_ &= ~OPTION_NOEVENTS;
  puts("");
}

void MidiDump::meta(int what, int len, unsigned char* data)
{
}

void MidiDump::program(int channel, int prg)
{
}

void MidiDump::control(int channel, int what, int value)
{
}

void MidiDump::noteon(int channel, int note, int vel)
{
  if (!dumptrack[trackno_-1] || !dumpchannel[channel])
    return;
  for (int i = 0; i < notecount_; i++)
  if (tactnote_[i].noteval_ == note && tactnote_[i].until_ < 0)
    return;
  if (notecount_ >= MAXNOTESPERTACT)
  {
    for (i = 0; i < notecount_; i++)
      if (tactnote_[i].noteval_ < 0)
	break;
    if (i >= notecount_)
      return;  // sorry for not dumping this note!
  }
  else
    i = notecount_++;

  tactnote_[i].noteval_ = note;
  tactnote_[i].vel_ = vel;
  tactnote_[i].since_ = curunit_;
  tactnote_[i].until_ = -1; // not finished

  if (curunit_ >= nexttacttime)
    dumptact(this);
}

void MidiDump::noteoff(int channel, int note, int vel)
{
  if (!dumptrack[trackno_-1] || !dumpchannel[channel])
    return;
  // find note
  for (int i = 0; i < notecount_; i++)
    if (tactnote_[i].noteval_ == note && tactnote_[i].until_ < 0)
    {
      tactnote_[i].until_ = curunit_;
      break;
    }
  if (curunit_ >= nexttacttime)
    dumptact(this);
}

void MidiDump::pitchbend(int channel, int val)
{
}

void MidiDump::polyaftertouch(int channel, int note, int val)
{
}

void MidiDump::aftertouch(int channel, int val)
{
}

void MidiDump::songpos(unsigned pos)
{
}

void MidiDump::songselect(unsigned char song)
{
}

void MidiDump::tunerequest()
{
}

void MidiDump::timingclock()
{
}

void MidiDump::start()
{
}

void MidiDump::cont()
{
}

void MidiDump::stop()
{
}

void MidiDump::activesense()
{
}

void MidiDump::sysex(int syslen, unsigned char* sysdata)
{
}

void MidiDump::error(const char* msg)
{
  printf("// error: %s\n", msg);
  exit(1);
}

char* notename[12] = {
  "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#","B"
};

const char* delay2str(long delay)
{
static char str[80];
char* s = str;

  if (delay == unitsperquarter*4)
  {
    strcpy(str, "4/4");
    delay = 0;
  }
  if (delay > unitsperquarter*4)
  {
    sprintf(s, "%ld", delay / (unitsperquarter*4));
    delay %= (unitsperquarter*4);
    s += strlen(s);
    if (delay > 0)
      *s++ = ' ';
  }

  if (delay > 0)
  {
    for (int base = 1; base <= 32; base *= 2)
    if (((delay * base) % (4 * unitsperquarter)) == 0)
    {
      sprintf(s, "%ld/%d", delay * base / (unitsperquarter*4), base);
      break;
    }
    if (base > 32)
      sprintf(s, "%ld units", delay);
  }
  return (const char*)str;
}

void dumpmeasure()
{
  printf("M%03ld\t", measure);
}

void sortevents(MidiDump* track)
{
  if (track->notecount_ < 2)
    return;
  for (int i = 0; i < track->notecount_; i++)
  {
  int bestj = -1;

    for (int j = i; j < track->notecount_; j++)
    if (track->tactnote_[j].noteval_ >= 0)
    {
      if (bestj < 0 || track->tactnote_[j].since_ < track->tactnote_[bestj].since_)
	bestj = j;
      else if (track->tactnote_[j].since_ == track->tactnote_[bestj].since_)
      {
	if ((track->tactnote_[j].until_ > track->tactnote_[bestj].until_))
	  bestj = j;
	else if (track->tactnote_[j].until_ == track->tactnote_[bestj].until_)
	{
	  if (track->tactnote_[bestj].noteval_ < track->tactnote_[j].noteval_)
	    bestj = j;
	}
      }
    }
    if (bestj < 0)  //end
      break;
    if (i == bestj)
      continue;
    if (track->tactnote_[i].noteval_ < 0)
    {
      track->tactnote_[i] = track->tactnote_[bestj];
      track->tactnote_[bestj].noteval_ = -1;
    }
    else
    {
      Note temp = track->tactnote_[bestj];
      track->tactnote_[bestj] = track->tactnote_[i];
      track->tactnote_[i] = temp;
    }
  }
}

void dumptact(MidiDump* track)
{
  dumpmeasure();
  printf("======= Beat: %s\n", delay2str(curtactdelay));

  sortevents(track);
  for (int i = 0; i < track->notecount_; i++)
  if (track->tactnote_[i].noteval_ >= 0 &&
      track->tactnote_[i].since_ < curtacttime + curtactdelay)
  {
     dumpmeasure();
     // start time
     if (curtacttime > track->tactnote_[i].since_)
       printf("TIE  ");
     else
     {
       int at = (int)(track->tactnote_[i].since_ - curtacttime);

       printf("%d.%03d ", at / unitsperquarter + 1, at % unitsperquarter);
     }

     printf(" %2s%d V%03d",
      notename[track->tactnote_[i].noteval_ % 12],
      track->tactnote_[i].noteval_ / 12 - 1,  // octave
      track->tactnote_[i].vel_
     );


     long delay;

      if (track->tactnote_[i].until_ < 0 || track->tactnote_[i].until_ > curtacttime + curtactdelay)
	printf("   TIE");
      else
      {
	if (track->tactnote_[i].since_ >= curtacttime)
	   delay = track->tactnote_[i].until_ - track->tactnote_[i].since_;
	else
	   delay = track->tactnote_[i].until_ - curtacttime; // tied delay
	printf("   %d.%03d", (int)(delay / unitsperquarter), (int)(delay % unitsperquarter));
      }
      printf("\n");
  }


  // remove notes that are finished:
  for (i = 0; i < track->notecount_; i++)
  if (track->tactnote_[i].noteval_ >= 0 &&
      track->tactnote_[i].until_ >= 0 &&
      track->tactnote_[i].until_ <= nexttacttime)
    track->tactnote_[i].noteval_ = -1;

  nexttact();
}

void mididump()
{
  MidiScan mainmidi(input);
  if (!mainmidi.getf())
  {
    perror(input);
    exit(1);
  }
  mainmidi.run();

  printf("\n");

  assert(dumptrack != 0);

  int i;

  for (i = 0; i < 16; i++)
    if (dumpchannel[i])
      break;
  if (i >= 16)
  {
    for (i = 0; i < 16; i++)
      dumpchannel[i] = 1;
  }

  trackread = new MidiDump(input);
  assert(trackread);

  trackread->run();
  delete trackread; trackread = 0;
}

void usage()
{
  fprintf(stderr, "usage: mididmp [-track #] [-channel #] file.mid\n");
  fprintf(stderr, "dumps midi file note events\n");
  exit(1);
}

int main(int argc, char** argv)
{
  argc--; argv++;

  while (argc > 0 && **argv == '-')
  {
    if (strncmp(*argv, "-track", 2) == 0)
    {
      argc--; argv++;
      if (argc == 0 || !isdigit(**argv))
	showtracks = 1;
      else
      {
	while (argc > 0 && isdigit(**argv))
	{
	int trackno = atoi(*argv);

	  if (trackno <= 0)
	    fprintf(stderr, "invalid track number %s\n", *argv);
	  else
	  {
	    if (option_trackcnt < MAXTRACKS)
	      option_track[option_trackcnt++] = trackno;
	    else
	    {
	      fprintf(stderr, "too many tracks: track %s ignored\n", *argv);
	    }
	  }
	  argc--; argv++;
	}
      }
      continue;
    }
    if (strncmp(*argv, "-channel", 2) == 0)
    {
      argc--; argv++;
      if (argc == 0 || !isdigit(**argv))
	showtracks = 1;
      else
      {
	while (argc > 0 && isdigit(**argv))
	{
	int channelno = atoi(*argv);

	  if (channelno <= 0 || channelno > 16)
	    fprintf(stderr, "invalid channel number %s (should be 1-16)\n", *argv);
	  else
	    dumpchannel[channelno-1] = 1;
	  argc--; argv++;
	}
      }
      continue;
    }

    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }

  if (argc < 1)
    usage();

  input = argv[0];


  if (showtracks)
  {
    MidiTrackinfo info(input);
    if (info.getf())
    {
      printf("File %s:\n", input);
      info.print(stdout);
      return 0;
    }
    else
    {
      perror(input);
      return 1;
    }
  }
  mididump();

  return 0;
}