// midifade written by Gnter Nagler 1995 (gnagler@ihm.tu-graz.ac.at)
#include "midiio.hpp"
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#ifdef __MSDOS__
#include <alloc.h>
#endif

class MidiFade;

char* input = 0;
char* output = 0;

#define UNDEFINED ((unsigned long)-1)

typedef struct
{
  unsigned long at;
  long microsecperbeat;
} TEMPOCHANGE;

typedef struct
{
  int trackno;
  int channel;
  unsigned long start, end;
} PLAYPART;

int startvel = 0, endvel = 100;  // percentage of original velocity
long fadestart = 0, fadeend = 1000;
int dir = 1;
int fadechannel[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
int fadeall = 1;

char* starttime = 0;
char* endtime = 0;
int n = 0;
#define MAXVEL 256  // maximum velocity percentage increase  (256% of original)
unsigned long volumepos[MAXVEL];

MidiWrite* write = 0;

int showtracktime = 0;
int showfulltime = 0;

void usage();

class MidiPlayTime : public MidiRead
{
public:
  MidiPlayTime(char* name);
  ~MidiPlayTime();

  virtual void head(unsigned version, unsigned tracks, unsigned unitsperbeat);
  virtual void track(int trackno, long length, int channel);
  virtual void endtrack(int trackno);

  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void error(const char* msg);

  virtual void endmidi();
  virtual void tempo(unsigned long ticks);

  unsigned long firstnoteon_;
  unsigned long lastnoteoff_;

  long minpause_; // min pause (units) to be a long break

  unsigned long microsec(unsigned long units);
  unsigned long units(unsigned long microsec);

  void showplay();  // when are the notes played?

protected:
  long note_[128];
  unsigned long recentnoteoff_, recentbreak_, firstnoteonafterbreak_;

  void addtempochange(long pos, long microsecperbeat);
  void sorttempochanges(); // sort by pos

  TEMPOCHANGE *tempochange_;
  int tempochanges_, maxtempochanges_;

  void addplay(int trackno, int channel, long start, long end);
  PLAYPART *play_;
  int plays_, maxplays_;

  int ismute();
};

unsigned long MidiPlayTime::units(unsigned long microsec)
{
unsigned long ms = 0;
unsigned long u;
int i;
long curtempo = 500000L;
long d, m;

  assert(unitsperbeat_ != 0); // call run() first!

  for (u = 0, i = 0; i < tempochanges_ && ms < microsec; i++)
  {
    d = tempochange_[i].at - u;
    if (d > 0)
    {
      m = MidiRead::microsec(d, curtempo);
      if (ms + m <= microsec)
      {
	u += d;
	ms += m;
      }
      else
      {
	// m microseconds with tempo curtempo
	u += MidiRead::units(microsec-ms, curtempo);
	ms = microsec;
      }
    }
    curtempo = tempochange_[i].microsecperbeat;
  }
  if (ms < microsec)
    u += MidiRead::units(microsec-ms, curtempo);
  return u;
}

unsigned long MidiPlayTime::microsec(unsigned long units)
{
unsigned long ms = 0;
long u;
int i;
long curtempo = 500000L;
long d;

  assert(unitsperbeat_ != 0); // call run() first!

  for (u = 0, i = 0; i < tempochanges_ && u < units; i++)
  {
    d = tempochange_[i].at - u;
    if (d > units-u)
      d = units-u;
    if (d > 0)
    {
      // d units with tempo curtempo
      ms += MidiRead::microsec(d, curtempo);
      u += d;
    }
    curtempo = tempochange_[i].microsecperbeat;
  }
  if (u < units)
    ms += MidiRead::microsec(units-u, curtempo);

  return ms;
}

#define MINUTE(ms) ((int)((ms) / 60000000l))
#define SECOND(ms) ((int)(((ms) / 1000000l)) % 60)
#define MICROSEC(ms)   ((ms) % 1000000l)

void MidiPlayTime::showplay()
{
unsigned long s, e;

  if (firstnoteon_ == UNDEFINED || lastnoteoff_ == UNDEFINED)
    printf("No notes played!\n");
  else
  {
    s = microsec(firstnoteon_);
    e = microsec(lastnoteoff_);
    printf("song:      	       units %6lu - %6lu   time %d:%02d.%06lu - %d:%02d.%06lu\n",
      firstnoteon_, lastnoteoff_,
      MINUTE(s), SECOND(s), MICROSEC(s),
      MINUTE(e), SECOND(e), MICROSEC(e));
  }

  if (!showtracktime)
    return;

  for (int i = 0; i < plays_; i++)
  {
    s = microsec(play_[i].start);
    e = microsec(play_[i].end);
    printf("track %-2d", play_[i].trackno);
    if (play_[i].channel >= 0)
      printf(" (channel %2d)", play_[i].channel+1);
    else
      printf("            ");
    printf(": units %6lu - %6lu   time %d:%02d.%06lu - %d:%02d.%06lu\n",
      play_[i].start, play_[i].end,
      MINUTE(s), SECOND(s), MICROSEC(s),
      MINUTE(e), SECOND(e), MICROSEC(e));
  }
}

void MidiPlayTime::addtempochange(long pos, long microsecperbeat)
{
  for (int i = 0; i < tempochanges_; i++)
  if (tempochange_[i].at == pos)
  {
    tempochange_[i].microsecperbeat = microsecperbeat;
    return;
  }
  if (tempochanges_ == maxtempochanges_)
  {
    TEMPOCHANGE* p;

    maxtempochanges_ += 10;
    if (tempochanges_ == 0)
      p = (TEMPOCHANGE*)malloc(sizeof(TEMPOCHANGE) * maxtempochanges_);
    else
      p = (TEMPOCHANGE*)realloc(tempochange_, sizeof(TEMPOCHANGE) * maxtempochanges_);
    if (p)
      tempochange_ = p;
    else
    {
      fprintf(stderr, "Error: out of memory\n");
      exit(1);
    }
  }
  if (tempochanges_ < maxtempochanges_)
  {
    tempochange_[tempochanges_].at = pos;
    tempochange_[tempochanges_].microsecperbeat = microsecperbeat;
    tempochanges_++;
  }
}

void MidiPlayTime::endmidi()
{
  sorttempochanges();
}

void MidiPlayTime::sorttempochanges()
{
int changed;

  // simple bubble sort
  do
  {
    changed = 0;
    for (int i = 0; i < tempochanges_-1; i++)
    if (tempochange_[i].at > tempochange_[i+1].at)
    {
     TEMPOCHANGE temp;

      temp = tempochange_[i];
      tempochange_[i] = tempochange_[i+1];
      tempochange_[i+1] = temp;
      changed = 1;
    }
  } while (changed);
}

void MidiPlayTime::addplay(int trackno, int channel, long start, long end)
{
  if (firstnoteon_ == UNDEFINED || start < firstnoteon_)
    firstnoteon_ = start;
  if (lastnoteoff_ == UNDEFINED || end > lastnoteoff_)
    lastnoteoff_ = end;

  if (plays_ == maxplays_)
  {
    PLAYPART* p;

    maxplays_ += 10;
    if (plays_ == 0)
      p = (PLAYPART*)malloc(sizeof(PLAYPART) * maxplays_);
    else
      p = (PLAYPART*)realloc(play_, sizeof(PLAYPART) * maxplays_);
    if (p)
      play_ = p;
    else
    {
      maxplays_ -= 10;  // not enough memory
      fprintf(stderr, "Warning: out of memory\n");
      return;
    }
  }
  if (plays_ < maxplays_)
  {
    play_[plays_].trackno = trackno;
    play_[plays_].channel = channel;
    play_[plays_].start = start;
    play_[plays_].end = end;
    plays_++;
  }
}

int MidiPlayTime::ismute()
{
  for (int i = 0; i < 128; i++)
    if (note_[i] != -1)
      return 0;
  return 1;
}

MidiPlayTime::MidiPlayTime(char* name)  : MidiRead(name)
{
  maxtempochanges_ = tempochanges_ = 0;
  tempochange_ = 0;
  addtempochange(0, 500000L); // 120 beats/minute initialization

  maxplays_ = plays_ = 0;
  play_ = 0;

  firstnoteon_ = lastnoteoff_ = UNDEFINED;
}

MidiPlayTime::~MidiPlayTime()
{
  if (tempochanges_ > 0)
  {
    free(tempochange_);
    tempochange_ = 0;
    tempochanges_ = 0;
  }
  if (plays_ > 0)
  {
    free(play_);
    play_ = 0;
    plays_ = 0;
  }
}

void MidiPlayTime::head(unsigned version, unsigned tracks, unsigned unitsperbeat)
{
  minpause_ = long(unitsperbeat_) * 4 * 5;
}

void MidiPlayTime::track(int trackno, long length, int channel)
{

  for (int i = 0; i < 128; i++)
    note_[i] = -1;  // mute
  recentnoteoff_ = recentbreak_ = 0;
  firstnoteonafterbreak_ = UNDEFINED;
}

void MidiPlayTime::endtrack(int trackno)
{
  if (firstnoteonafterbreak_ != UNDEFINED)
     addplay(trackno_, curchannel_, firstnoteonafterbreak_, recentnoteoff_);
}

void MidiPlayTime::noteon(int channel, int note, int vel)
{
  if (curunit_ - recentnoteoff_ >= minpause_ && ismute())
  {
    // a long break
    if (firstnoteonafterbreak_ != UNDEFINED)
      addplay(trackno_, curchannel_, firstnoteonafterbreak_, recentnoteoff_);

    firstnoteonafterbreak_ = curunit_;
  }
  if (firstnoteonafterbreak_ == UNDEFINED)
    firstnoteonafterbreak_ = curunit_;
  note_[note] = curunit_;
}

void MidiPlayTime::noteoff(int channel, int note, int vel)
{
  recentnoteoff_ = curunit_;
  note_[note] = -1;
}

void MidiPlayTime::error(const char* msg)
{
  fprintf(stderr, "Error: %s\n", msg);
  exit(1);
}

void MidiPlayTime::tempo(unsigned long microsecperbeat)
{
  addtempochange(curunit_, microsecperbeat);
}


class MidiFade : public MidiRead
{
public:
  MidiFade(char* name, FILE* f = 0);

  virtual void head(unsigned version, unsigned tracks, unsigned unitsperbeat);
  virtual void track(int trackno, long length, int channel);
  virtual void endtrack(int trackno);

  virtual void program(int prg, int channel);
  virtual void control(int channel, int what, int value);
  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void pitchbend(int channel, int val);
  virtual void polyaftertouch(int channel, int note, int val);
  virtual void aftertouch(int channel, int val);
  virtual void songpos(unsigned pos);
  virtual void songselect(unsigned char song);
  virtual void tunerequest();
  virtual void timingclock();
  virtual void start();
  virtual void cont();
  virtual void stop();
  virtual void activesense();
  virtual void sysex(int syslen, unsigned char* sysdata);
  virtual void meta(int what, int len, unsigned char* data);
  virtual void time(unsigned long ticks);
  virtual void error(const char* msg);

  int seekvel(int defvel);
};

int MidiFade::seekvel(int defvel)
{
  if (curunit_ >= fadestart && curunit_ <= fadeend)
  {
    for (int i = startvel; i != endvel+dir; i+=dir)
    if (i >= 0 && i < MAXVEL)
    {
      if (volumepos[i] <= curunit_ && (startvel == endvel || curunit_ < volumepos[i+dir]))
      {
	int newvel = (int)((long(defvel) * i) / 100);  // i percent of defvel
	if (newvel < 0)
	  return 0;
	if (newvel > 127)
	  return 127;
//	fprintf(stderr, "%lu: changed velocity from %d to %d\n", curunit_, defvel, newvel);
	return newvel;
      }
    }
    return defvel;
  }
  else
    return defvel;
}


MidiFade::MidiFade(char* name, FILE* f) : MidiRead(name, f)
{
  options_ = OPTION_NOCONTROLS+OPTION_NOSYSEVENTS;
}

void MidiFade::head(unsigned version, unsigned tracks, unsigned unitsperbeat)
{
  write->head(version, tracks, unitsperbeat);
}

void MidiFade::track(int trackno, long length, int channel)
{
  write->track();
}

void MidiFade::endtrack(int trackno)
{
  write->endtrack();
}


void MidiFade::time(unsigned long ticks)
{
  write->time(ticks);
}

void MidiFade::meta(int what, int len, unsigned char* data)
{
  write->meta(what, len, data);
}

void MidiFade::program(int channel, int prg)
{
  write->program(channel, prg);
}

void MidiFade::control(int channel, int what, int value)
{
  write->control(channel, what, value);
}

void MidiFade::noteon(int channel, int note, int vel)
{
  write->noteon(channel, note, fadechannel[channel] ? seekvel(vel) : vel);
}

void MidiFade::noteoff(int channel, int note, int vel)
{
  write->noteoff(channel, note, fadechannel[channel] ? seekvel(vel) : vel);
}

void MidiFade::pitchbend(int channel, int val)
{
  write->pitchbend(channel, val);
}

void MidiFade::polyaftertouch(int channel, int note, int val)
{
  write->polyaftertouch(channel, note, val);
}

void MidiFade::aftertouch(int channel, int val)
{
  write->aftertouch(channel, val);
}

void MidiFade::songpos(unsigned pos)
{
  write->songpos(pos);
}

void MidiFade::songselect(unsigned char song)
{
  write->songselect(song);
}

void MidiFade::tunerequest()
{
  write->tunerequest();
}

void MidiFade::timingclock()
{
  write->timingclock();
}

void MidiFade::start()
{
  write->start();
}

void MidiFade::cont()
{
  write->cont();
}

void MidiFade::stop()
{
  write->stop();
}

void MidiFade::activesense()
{
  write->activesense();
}

void MidiFade::sysex(int syslen, unsigned char* sysdata)
{
  write->sysex(syslen, sysdata);
}

void MidiFade::error(const char* msg)
{
  printf("// error: %s\n", msg);
  exit(1);
}

// returns midi units
unsigned long gettime(MidiPlayTime& playtime, char* s)
{
unsigned long v = 0;

  if (!s)
    return v;
  while (isdigit(*s))
  {
    v = v * 10 + (*s - '0');
    s++;
  }
  if (*s == 'b' || *s == 'B')
    return v * playtime.unitsperbeat_;
  if (*s == 'u' || *s == 'U')
    return v;
  if (*s == ':')
  {
    // time in mm:ss.microsec
    unsigned long microsec = 0;

    if (v >= 70)
    {
      fprintf(stderr, "timeformat %s is too large (> 70 min)\n", s);
      v = 70;
    }
    microsec += v * 60000000UL;
    s++; // :
    if (!isdigit(*s))
    {
      fprintf(stderr, "invalid time format %s\n", s);
      usage();
    }
    v = 0;
    while (isdigit(*s))
    {
      v = v * 10 + (*s - '0');
      s++;
    }
    if (v >= 60)
    {
      fprintf(stderr, "time format second parameter is too large\n");
      v = 59;
    }
    microsec += v * 1000000l;

    if (*s == '.')
    {
      s++;
      v = 0;
      while (isdigit(*s))
      {
	v = v * 10 + (*s - '0');
	s++;
      }
      if (v >= 1000000l)
      {
	fprintf(stderr, "time format microsecond parameter too large\n");
	v = 999999l;
      }
      microsec += v;
    }
    return playtime.units(microsec);
  }
  fprintf(stderr, "invalid time format %s\n", s);
  usage();
  return 0;
}

void midifade()
{
  MidiPlayTime playtime(input);
  if (!playtime.getf())
  {
    perror(input);
    return;
  }
  playtime.run();
  if (showtracktime || showfulltime)
    playtime.showplay();
  else
  {
    fadestart = gettime(playtime, starttime);
    if (!endtime)
      fadeend = playtime.lastnoteoff_;
    else
      fadeend = gettime(playtime, endtime);
    if (fadeend < 0 || fadestart >= fadeend)
    {
      printf("invalid time range specified\n");
      usage();
      return;
    }
    if (fadeall)
    {
      for (int i = 0; i < 16; i++)
	fadechannel[i] = 1;
    }

    fprintf(stderr, "fade %schannels", fadeall ? "all " : "");
    if (!fadeall)
    {
      for (int i = 0; i < 16; i++)
	if (fadechannel[i])
	  fprintf(stderr, " %d", i+1);
    }
    fprintf(stderr, " between unit %lu and unit %lu from velocity %d%% to %d%%\n",
      fadestart, fadeend, startvel, endvel);

    dir = (endvel > startvel) ? +1 : -1;

    n = (endvel - startvel + 1) * dir;

    for (int vel = startvel; vel != endvel+dir; vel += dir)
    if (vel >= 0 && vel < MAXVEL)
    {
      if (endvel == startvel)
	volumepos[vel] = fadestart;
      else
	volumepos[vel] = fadestart + ((fadeend - fadestart) * (vel-startvel)) / (endvel - startvel);
    }

    MidiFade fade(input);
    if (!fade.getf())
    {
      perror(input);
      return;
    }
    write = new MidiWrite(output);
    if (!write)
    {
      fprintf(stderr, "out of memory\n");
      return;
    }
    if (!write->getf())
    {
      perror(output);
      return;
    }

    fade.run();

    delete write;  // close midi file
    fprintf(stderr, "output written to %s\n", output);
  }
}

void usage()
{
  fprintf(stderr, "midifade by G. Nagler changes linear velocity\n");
  fprintf(stderr, "usage: midifade [options] srcfile.mid dstfile.mid\n");
  fprintf(stderr, "-time\t\tshow time table where notes are played\n");
  fprintf(stderr, "-songtime\tshow time of whole song only\n");
  fprintf(stderr, "-channel # #\tfade only these channels # # (1-16)\n");
  fprintf(stderr, "-sv #\t\tstart velocity percentage (0-100)\n");
  fprintf(stderr, "-ev #\t\tend velocity percentage (0-100)\n");
  fprintf(stderr, "-st #\t\tstart time\n");
  fprintf(stderr, "-et #\t\tend time\n");
  fprintf(stderr, "-fadein\t\tsame as -sv 0 -ev 100 -st 0:0\n");
  fprintf(stderr, "-fadeout\t\tsame as -sv 100 -ev 0 no -et (end of song)\n");
  fprintf(stderr, "accepted time formats are:\n");
  fprintf(stderr, "  mm:ss[.microseconds] or #u in midi units or #b in beats\n");
  exit(1);
}

int main(int argc, char** argv)
{
  argc--; argv++;

  while (argc > 0 && **argv == '-')
  {

    if (strncmp(*argv, "-channels", 2) == 0)
    {
      argv++; argc--;
      if (argc > 0 && !isdigit(**argv))
      {
	fprintf(stderr, "option -channels needs one or more channel numbers (1-16)\n");
	continue;
      }
      while (argc > 0 && isdigit(**argv))
      {
	int channel = atoi(*argv);
	if (channel >= 1 && channel <= 16)
	{
	  fadeall = 0;
	  fadechannel[channel-1] = 1;
	}
	else
	  fprintf(stderr, "invalid channel number: %s (should be 1-16)\n", *argv);
	argc--; argv++;
      }
      continue;
    }
    if (strncmp(*argv, "-time", 2) == 0)
    {
      showfulltime = 1;
      showtracktime = 1;
      argc--; argv++; continue;
    }

    if (strncmp(*argv, "-songtime", 3) == 0)
    {
      showtracktime = 0;
      showfulltime = 1;
      argc--; argv++; continue;
    }

    if (strcmp(*argv, "-sv") == 0)
    {
      argc--; argv++;
      if (argc == 0 || !isdigit(**argv))
	fprintf(stderr, "option -sv needs start velocity percentage (0-100)\n");
      else
      {
	startvel = atoi(*argv);
	if (startvel < 0)
	{
	  startvel = 0;
	  fprintf(stderr, "invalid start velocity value %s (should be 0-100)\n", *argv);
	}
	argc--; argv++; continue;
      }
    }
    if (strcmp(*argv, "-ev") == 0)
    {
      argc--; argv++;
      if (argc == 0 || !isdigit(**argv))
	fprintf(stderr, "option -ev needs end velocity value (0-100)\n");
      else
      {
	endvel = atoi(*argv);
	if (endvel < 0)
	{
	  endvel = 0;
	  fprintf(stderr, "invalid end velocity value %s (should be 0-100)\n", *argv);
	}
	argc--; argv++; continue;
      }
    }
    if (strncmp(*argv, "-fadein", 6) == 0)
    {
      startvel = 0;
      endvel = 100;
      starttime = "0:0";
      argc--; argv++; continue;
    }
    if (strncmp(*argv, "-fadeout", 6) == 0)
    {
      startvel = 100;
      endvel = 0;
      endtime = NULL; // use default end time
      argc--; argv++; continue;
    }
    if (strcmp(*argv, "-st") == 0)
    {
      argc--; argv++;
      if (argc == 0 || !isdigit(**argv))
      {
	fprintf(stderr, "option -st needs start time value in a timeformat\n");
	usage();
      }
      else
      {
	starttime = *argv++; argc--;
      }
      continue;
    }
    if (strcmp(*argv, "-et") == 0)
    {
      argc--; argv++;
      if (argc == 0 || !isdigit(**argv))
      {
	fprintf(stderr, "option -et needs end time value in a timeformat\n");
	usage();
      }
      else
      {
	endtime = *argv++; argc--;
      }
      continue;
    }
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  if (argc < ((showtracktime || showfulltime) ? 1 : 2))
    usage();

  input = argv[0];
  if (!showtracktime && !showfulltime)
  {
    output = argv[1];
    if (strcmp(input, output) == 0)
    {
      fprintf(stderr, "cannot convert midi to same file\n");
      return 1;
    }
  }

  midifade();
  return 0;
}