// midigrab v1.0 written by Gnter Nagler 1995 (gnagler@ihm.tu-graz.ac.at)
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef __MSDOS__
#define READWRITE_BINARY  "r+b"
#define CREATE_READWRITE_BINARY  "w+b"
#define WRITE_BINARY  "wb"
#define READ_BINARY   "rb"
#else
#define READWRITE_BINARY  "r+"
#define CREATE_READWRITE_BINARY  "w+"
#define WRITE_BINARY  "w"
#define READ_BINARY   "r"
#endif

char* prefix = "grab";
char midiname[80];
static int saved = 0;

long getlong(FILE* f)
{
unsigned char v[4];
long len;

  if (fread(v, 1, 4, f) != 4)
    return -1;
  len = v[0];
  len = (len << 8) + v[1];
  len = (len << 8) + v[2];
  len = (len << 8) + v[3];
  return len;
}

unsigned int getword(FILE* f)
{
unsigned char v[2];
unsigned int len = 0;

  if (fread(v, 1, 2, f) == 2)
  {
    len = v[0];
    len = (len << 8) + v[1];
  }
  return len;
}

long findbytes(FILE* f, unsigned char *s, int len)
{
long oldpos = ftell(f);
long pos = -1, markpos;
int c, matchlen;

  if (s && len > 0)
  {
    while ((c = fgetc(f)) != EOF)
    {
      if (c == *s)
      {
	markpos = ftell(f);
	matchlen = 1;
	while (matchlen < len)
	{
	  c = fgetc(f);
	  if (c == EOF)
	    break;
	  if (c == s[matchlen])
	    matchlen++;
	  else
	    break;
	}
	if (matchlen == len)
	{
	  pos = markpos-1;
	  break;
	}
	else
	  fseek(f, markpos, SEEK_SET);
      }
    }
  }
  fseek(f, oldpos, SEEK_SET);
  return pos;
}

long findmidi(FILE* f)
{
unsigned char head[] = "MThd\x00\x00\x00\x06";

  return findbytes(f, head, 8);
}

long findtrack(FILE* f)
{
unsigned char trk[] = "MTrk";

  return findbytes(f, trk, 4);
}

int copybytes(FILE* df, FILE* sf, long n)
{
  while (n-- > 0)
  {
    int c = fgetc(sf);
    if (c < 0)
    {
      perror(midiname);
      return 0;
    }
    else
      fputc(c, df);
  }
  return 1;
}

int savemidi(FILE* f, long midistart)
{
  fseek(f, midistart, SEEK_SET); // get information
  getlong(f); // MThd
  if (getlong(f) != 6)
    return 0; // len=6
  unsigned int version = getword(f);
  if (version > 2)
    return 0; // invalid version
  unsigned int trackcount = getword(f);
  if (trackcount <= 0)
    return 0;
  getword(f); // resolution
  for (unsigned int i = 0; i < trackcount; i++)
  {
    long trackpos = ftell(f);
    if (findtrack(f) != trackpos)
      return 0;
    getlong(f); // MTrk
    long tracklen = getlong(f);
    fseek(f, tracklen, SEEK_CUR);
  }
  // ok
  long midiend = ftell(f);
  fseek(f, midistart, SEEK_SET);

  do
  {
    sprintf(midiname, "%s%d.mid", prefix, ++saved);

    FILE* testf = fopen(midiname, "r");
    if (!testf)
      break;
    fclose(testf);
  } while (1);

  printf("midi found: writing to %s\n", midiname);
  FILE* midif = fopen(midiname, CREATE_READWRITE_BINARY);
  if (!midif)
  {
    perror(midiname);
    return 0;
  }
  if (!copybytes(midif, f, midiend-midistart))
  {
    fclose(midif);
    return 0;
  }
  fclose(midif);
  fseek(f, midistart, SEEK_SET);
  return 1;
}

int midigrab(FILE* f)
{
long midistart;

   while ((midistart = findmidi(f)) >= 0)
   {
     savemidi(f, midistart);
     fseek(f, midistart+8, SEEK_SET);
   }
   return 1;
}

int midigrab(char* filename)
{
  FILE* f = fopen(filename, READ_BINARY);
  if (!f)
  {
    perror(filename);
    return 1;
  }

  int ret = midigrab(f);

  fclose(f);
  return ret;
}

int main(int argc, char** argv)
{
  argc--; argv++;
  if (!argc)
  {
    fprintf(stderr, "usage: midigrab filename ...\n");
    return 1;
  }
  while (argc > 0)
  {
    midigrab(*argv);
    argc--;
  }
  if (saved == 0)
    fprintf(stderr, "Sorry, no valid standard midi found.\n");
  return 0;
}

