#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "midiio.hpp"


static char* version = "midistat v1.0 by Gnter Nagler (" __DATE__ ")";

int notes[128];
int programs[128];
int channels[16];
int velocities[128];
int volumes[128];

int listcount(int* list, int n)
{
int count = 0;

  for (int i = 0; i < n; i++)
  if (list[i])
    count++;
  return count;
}

void printitem(int* list, int i)
{
  if (list == notes)
    printf("%s", MidiRead::notename(i));
  else if (list == programs)
    printf("%s", MidiRead::progname(i, 0)); // not drum
  else if (list == channels)
    printf("%d", SAYCHANNEL(i));
  else
    printf("%d", i);
}

void printlist(int* list, int count, int range = 1)
{
  int v = 0;
  int firstv = -1;
  while (v <= count)
  {
    if (firstv < 0 && v >= count)
      break;
    if (!range && list[v])
      firstv = v++;
    if (firstv >= 0 && (v >= count || !list[v] || !range))
    {
      if (firstv == v-2)
	v--;
      if (firstv == v-1)
      {
	putchar(' ');
	printitem(list, firstv);
      }
      else
      {
	putchar(' ');
	printitem(list, firstv);
	putchar('-');
	printitem(list, v-1);
      }
      firstv = -1;
      if (v >= count)
	break;
      continue;
    }
    if (firstv < 0 && !list[v])
    {
      v++;
      continue;
    }
    if (firstv >= 0 && list[v])
    {
      v++;
      continue;
    }
    if (firstv < 0 && list[v])
    {
      firstv = v++;
      continue;
    }
  }
  putchar('\n');
}

class MidiStatistic : public MidiRead
{
public:
  MidiStatistic(const char* filename);

  virtual void head(unsigned version, unsigned tracks, unsigned unitsperbeat);
  virtual void track(int trackno, long length, int channel);
  virtual void endtrack(int trackno);

  virtual void program(int channel, int prg);
  virtual void control(int channel, int what, int val);
  virtual void highbank(int channel, int val);
  virtual void wheel(int channel, int val);
  virtual void breath(int channel, int val);
  virtual void foot(int channel, int val);
  virtual void portamentotime(int channel, int val);
  virtual void data(int channel, int val);
  virtual void volume(int channel, int val);
  virtual void balance(int channel, int val);
  virtual void expression(int channel, int val);
  virtual void lowbank(int channel, int val);
  virtual void hold(int channel, int val);
  virtual void reverb(int channel, int val);
  virtual void chorus(int channel, int val);
  virtual void datainc(int channel, int val);
  virtual void datadec(int channel, int val);
  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void pitchbend(int channel, int val);
  virtual void polyaftertouch(int channel, int note, int val);
  virtual void aftertouch(int channel, int val);

};

MidiStatistic::MidiStatistic(const char* filename) : MidiRead(filename)
{
  options_ = OPTION_NOMETAEVENTS+OPTION_NOSYSEVENTS;
}

void MidiStatistic::head(unsigned version, unsigned tracks, unsigned unitsperbeat)
{
  printf("%s: %d Track%s\n", midiname_, tracks, (tracks == 1 ? "" : "s"));
}

void MidiStatistic::track(int trackno, long length, int channel)
{
  for (int i = 0; i  < 16;i++)
    channels[i] = 0;

  for (i = 0; i < 128; i++)
  {
    velocities[i] = 0;
    volumes[i] = 0;
    notes[i] = 0;
    programs[i] = 0;
  }
}

void MidiStatistic::endtrack(int trackno)
{
  printf("Track %d:\n", trackno);
  if (listcount(channels, 16) > 0)
  {
    printf("  Channels:");
    printlist(channels, 16);
  }
  if (listcount(notes, 128) > 0)
  {
    printf("  Notes:");
    printlist(notes, 128);
  }
  if (listcount(programs, 128) > 0)
  {
    printf("  Programs:");
    printlist(programs, 128, 0);
  }

  if (listcount(velocities, 128) > 0)
  {
    printf("  Velocity:");
    printlist(velocities, 128);
  }

  if (listcount(volumes, 128) > 0)
  {
    printf("  Volume:");
    printlist(volumes, 128);
  }
}

void MidiStatistic::program(int channel, int program)
{
  if (channel != gm_drumchannel)
    programs[program] = 1;
  channels[channel] = 1;
}

void MidiStatistic::control(int channel, int what, int value)
{
  channels[channel] = 1;
}

void MidiStatistic::highbank(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::wheel(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::breath(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::foot(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::portamentotime(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::data(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::volume(int channel, int val)
{
  volumes[val] = 1;
  channels[channel] = 1;
}

void MidiStatistic::balance(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::expression(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::lowbank(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::hold(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::reverb(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::chorus(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::datainc(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::datadec(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::noteon(int channel, int note, int vel)
{
  notes[note] = 1;
  channels[channel] = 1;
  if (vel > 0)
    velocities[vel] = 1;
}

void MidiStatistic::noteoff(int channel, int note, int vel)
{
  notes[note] = 1;
  channels[channel] = 1;
}

void MidiStatistic::pitchbend(int channel, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::polyaftertouch(int channel, int note, int val)
{
  channels[channel] = 1;
}

void MidiStatistic::aftertouch(int channel, int val)
{
  channels[channel] = 1;
}


int chk(char* filename)
{
  MidiStatistic stat(filename);
  if (!stat.getf())
  {
    perror(filename);
    return 1;
  }
  return stat.run();
}

int main(int argc, char**argv)
{
  argc--; argv++;
  while (argc > 0 && **argv == '-')
  {
    if (strncmp(*argv, "-version", 2) == 0)
    {
      fprintf(stderr, "%s\n", version);
      argc--; argv++;
      if (argc == 0)
	return 0;
      continue;
    }
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
  }
  if (!argc)
  {
    fprintf(stderr, "usage: midistat filename.mid [> output]\n");
    return 1;
  }

  return chk(*argv);
}