// miditran v1.0 written by Gnter Nagler 1995 (gnagler@ihm.tu-graz.ac.at)
#include "midiio.hpp"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef __MSDOS__
#define WRITE_BINARY  "wb"
#define READ_BINARY   "rb"
#else
#define WRITE_BINARY  "w"
#define READ_BINARY   "r"
#endif

char* input = 0;
char* output = 0;

MidiWrite* write = 0;
int quiet = 0;

int incrementnote = 0;
int incrementvel = 0;
int incrementvol = 0;

int incr_channel[16] = {
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0
};

class MidiTranspose : public MidiRead
{
public:
  MidiTranspose(char* name);

  virtual void head(unsigned version, unsigned tracks, unsigned unitsperbeat);
  virtual void track(int trackno, long length, int channel);
  virtual void endtrack(int trackno);

  virtual void text(int what, int len, char* whattext, unsigned char* txt);
  virtual void meta(int what, int len, unsigned char* data);
  virtual void end();
  virtual void tact(int nom, int denom, int v1, int v2);
  virtual void tempo(unsigned long ticks);
  virtual void program(int prg, int channel);
  virtual void control(int channel, int what, int value);
  virtual void highbank(int channel, int val);
  virtual void wheel(int channel, int val);
  virtual void breath(int channel, int val);
  virtual void foot(int channel, int val);
  virtual void portamentotime(int channel, int val);
  virtual void data(int channel, int val);
  virtual void volume(int channel, int val);
  virtual void balance(int channel, int val);
  virtual void expression(int channel, int val);
  virtual void lowbank(int channel, int val);
  virtual void hold(int channel, int val);
  virtual void reverb(int channel, int val);
  virtual void chorus(int channel, int val);
  virtual void datainc(int channel, int val);
  virtual void datadec(int channel, int val);
  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void time(unsigned long ticks);
  virtual void pitchbend(int channel, int val);
  virtual void polyaftertouch(int channel, int note, int val);
  virtual void aftertouch(int channel, int val);
  virtual void songpos(unsigned pos);
  virtual void songselect(unsigned char song);
  virtual void tunerequest();
  virtual void timingclock();
  virtual void start();
  virtual void cont();
  virtual void stop();
  virtual void activesense();
  virtual void sysex(int syslen, unsigned char* sysdata);
  virtual void error(const char* msg);

  virtual void percent(int perc);
};

MidiTranspose::MidiTranspose(char* name) : MidiRead(name)
{
}

void MidiTranspose::head(unsigned version, unsigned tracks, unsigned unitsperbeat)
{
  write->head(version, 0, unitsperbeat);
}

void MidiTranspose::track(int trackno, long length, int channel)
{
  write->track();
}

void MidiTranspose::endtrack(int trackno)
{
  write->endtrack();
}

void MidiTranspose::text(int what, int len, char* whattext, unsigned char* txt)
{
  write->text(what, len, txt);
}

void MidiTranspose::meta(int what, int len, unsigned char* data)
{
  write->meta(what, len, data);
}

void MidiTranspose::end()
{
}

void MidiTranspose::tact(int nom, int denom, int v1, int v2)
{
  write->tact(nom, denom, v1, v2);
}

void MidiTranspose::tempo(unsigned long ticks)
{
  write->tempo(ticks);
}

void MidiTranspose::time(unsigned long ticks)
{
  write->time(ticks);
}

void MidiTranspose::program(int channel, int prg)
{
  write->program(channel, prg);
}

void MidiTranspose::control(int channel, int what, int value)
{
  write->control(channel, what, value);
}

void MidiTranspose::volume(int channel, int val)
{
  if (incr_channel[channel])
  {
    val += incrementvol;
    if (val > 127)
      val = 127;
    if (val < 0)
      val = 0;
  }
  write->volume(channel, val);
}

void MidiTranspose::balance(int channel, int val)
{
  write->balance(channel, val);
}

void MidiTranspose::wheel(int channel, int val)
{
  write->wheel(channel, val);
}

void MidiTranspose::breath(int channel, int val)
{
  write->breath(channel, val);
}

void MidiTranspose::foot(int channel, int val)
{
  write->foot(channel, val);
}

void MidiTranspose::portamentotime(int channel, int val)
{
  write->portamentotime(channel, val);
}

void MidiTranspose::data(int channel, int val)
{
  write->data(channel, val);
}

void MidiTranspose::expression(int channel, int val)
{
  write->expression(channel, val);
}

void MidiTranspose::lowbank(int channel, int val)
{
  write->lowbank(channel, val);
}

void MidiTranspose::highbank(int channel, int val)
{
  write->highbank(channel, val);
}

void MidiTranspose::hold(int channel, int val)
{
  write->hold(channel, val);
}

void MidiTranspose::reverb(int channel, int val)
{
  write->reverb(channel, val);
}

void MidiTranspose::chorus(int channel, int val)
{
  write->chorus(channel, val);
}

void MidiTranspose::datainc(int channel, int val)
{
  write->datainc(channel, val);
}

void MidiTranspose::datadec(int channel, int val)
{
  write->datadec(channel, val);
}

void MidiTranspose::noteon(int channel, int note, int vel)
{
  if (incr_channel[channel])
  {
    vel += incrementvel;
    if (vel > 127)
      vel = 127;
    if (vel < 0)
      vel = 0;
    if (channel != gm_drumchannel)
    {
      note += incrementnote;
      if (note > 127)
	note = 127;
    if (note < 0)
      note = 0;
    }
  }
  write->noteon(channel, note, vel);
}

void MidiTranspose::noteoff(int channel, int note, int vel)
{
  if (incr_channel[channel])
  {
    vel += incrementvel;
    if (vel > 127)
      vel = 127;
    if (vel < 0)
      vel = 0;
    if (channel != gm_drumchannel)
    {
      note += incrementnote;
      if (note > 127)
	note = 127;
      if (note < 0)
	note = 0;
    }
  }
  write->noteoff(channel, note, vel);
}

void MidiTranspose::pitchbend(int channel, int val)
{
  write->pitchbend(channel, val);
}

void MidiTranspose::polyaftertouch(int channel, int note, int val)
{
  write->polyaftertouch(channel, note, val);
}

void MidiTranspose::aftertouch(int channel, int val)
{
  write->aftertouch(channel, val);
}

void MidiTranspose::songpos(unsigned pos)
{
  write->songpos(pos);
}

void MidiTranspose::songselect(unsigned char song)
{
  write->songselect(song);
}

void MidiTranspose::tunerequest()
{
  write->tunerequest();
}

void MidiTranspose::timingclock()
{
  write->timingclock();
}

void MidiTranspose::start()
{
  write->start();
}

void MidiTranspose::cont()
{
  write->cont();
}

void MidiTranspose::stop()
{
  write->stop();
}

void MidiTranspose::activesense()
{
  write->activesense();
}

void MidiTranspose::sysex(int syslen, unsigned char* sysdata)
{
  write->sysex(syslen, sysdata);
}

void MidiTranspose::percent(int perc)
{
  if (!quiet)
    fprintf(stderr, "%-3d%%\r", perc);
}

void MidiTranspose::error(const char* msg)
{
  printf("// error: %s\n", msg);
}

void usage()
{
  fprintf(stderr, "MidiTranspose increments notes,velocity,volume by a given value\n");
  fprintf(stderr, "usage: MidiTranspose [-q][-channel #][-velocity #][-volume #][-note #] file.mid transpos.mid\n");
  fprintf(stderr, "-q\tbe quiet\n");
  fprintf(stderr, "-velocity #\tadd # to all note velocities (-127..+127)\n");
  fprintf(stderr, "-volume #\tadd # to all volume events (-127..+127)\n");
  fprintf(stderr, "-note#\tadd # to all note values (-127..+127)\n");
  fprintf(stderr, "-channel #\tchanges are done only on these channels (1-16)\n");
  fprintf(stderr, "At least one of the options -velocity, -volume, -note must be used.\n");
  fprintf(stderr, "Default are all channels (except drum channel 10 when incrementing notes)\n");
  exit(1);
}

int isnumber(int c)
{
  return isdigit(c) || c == '+' || c == '-';
}

int main(int argc, char**argv)
{
  argc--; argv++;
  if (argc == 0)
    usage();
  while (argc > 0 && **argv == '-')
  {
    if (argc > 0 && strcmp(*argv, "-q") == 0)
    {
      quiet = 1;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-channel", 2) == 0)
    {
      argc--; argv++;
      if (argc <= 0 || !isdigit(**argv))
	fprintf(stderr, "option -channel needs channel numbers as parameter (1-16)\n");
      while (argc > 0 && isdigit(**argv))
      {
	int ch = atoi(*argv);
	if (ch < 1 || ch > 16)
	  fprintf(stderr, "invalid channel number %s (should be 1-16)\n", *argv);
	else
	  incr_channel[ch-1] = 1;
	argc--; argv++;
      }
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-velocity", 3) == 0)
    {
      argc--; argv++;
      if (argc <= 0 || !isnumber(**argv))
	fprintf(stderr, "option -velocity needs increment value (-127,+127)\n");
      else
      {
	incrementvel = atoi(*argv);
	argc--; argv++;
      }
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-volume", 3) == 0)
    {
      argc--; argv++;
      if (argc <= 0 || !isnumber(**argv))
	fprintf(stderr, "option -volume needs increment value (-127,+127)\n");
      else
      {
	incrementvol = atoi(*argv);
	argc--; argv++;
      }
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-note", 2) == 0)
    {
      argc--; argv++;
      if (argc <= 0 || !isnumber(**argv))
	fprintf(stderr, "option -note needs increment value (-127,+127)\n");
      else
      {
	incrementnote = atoi(*argv);
	argc--; argv++;
      }
      continue;
    }
     fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  if (!incrementvel && !incrementvol && !incrementnote)
  {
    fprintf(stderr, "one of the options -velocity or -volume or -note must be specified!\n");
    usage();
  }

  if (argc < 2)
    usage();

  input = argv[0];
  output = argv[1];
  if (strcmp(input, output) == 0)
  {
    fprintf(stderr, "cannot convert midi to same file\n");
    return 1;
  }

  for (int i = 0; i < 16; i++)
  if (incr_channel[i])
    break;
  if (i >= 16)
  {
    for (i = 0; i < 16; i++)
      incr_channel[i] = 1;
    fprintf(stderr, "Warning: option -channel not used. Transposing channels: ");
    if (incrementvol || incrementvel)
      printf("1-16.\n");
    else
    {
      printf("1-9 and 11-16.\n");
      incr_channel[9] = 0;
    }
  }

  MidiTranspose midi(input);
  if (!midi.getf())
  {
    perror(input);
    return 1;
  }
  write = new MidiWrite(output);
  if (!write)
  {
    fprintf(stderr, "out of memory\n");
    return 1;
  }
  if (!write->getf())
  {
    perror(output);
    return 1;
  }
  argc--; argv++;
  if (!midi.run())
    fprintf(stderr, "%s: midi read error at %04lX\n", input, midi.getpos());

  delete write;
  return 0;
}
