// midreset written by Gnter Nagler 1997 (gnagler@ihm.tu-graz.ac.at)
#include "midiio.hpp"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

static char* version = "midreset v1.1 by Gnter Nagler (" __DATE__ ")";

char* input = 0;
char* output = 0;

MidiWrite* write = 0;

int xg_reset = 0;
int gs_reset = 0;
int gm_reset = 0;

class MidiReset : public MidiCopy
{
public:
  MidiReset(char* name);

  virtual void track(int trackno, long length, int channel);

  virtual void gmreset();
  virtual void gsreset();
  virtual void xgreset();
  virtual void gsexit();

  virtual void error(const char* msg);
  virtual void warning(const char* msg);

  virtual void percent(int perc);

  void i3_initchannel(int channel);

};

MidiReset::MidiReset(char* name) : MidiCopy(name)
{
}


void MidiReset::track(int trackno, long length, int channel)
{
  MidiCopy::track(trackno, length, channel);
  if (trackno == 1)
  {
    if (gm_reset)
      MidiCopy::gmreset();
    if (xg_reset)
      MidiCopy::xgreset();
    if (gs_reset)
      MidiCopy::gsreset();
  }
}

void MidiReset::gmreset() // ignore
{
}

void MidiReset::gsreset() // ignore
{
}

void MidiReset::xgreset() // ignore
{
}

void MidiReset::gsexit()  // ignore
{
}

void MidiReset::percent(int perc)
{
  fprintf(stderr, "%-3d%%\r", perc);
}

void MidiReset::error(const char* msg)
{
  printf("Error: %s\n", msg);
  exit(1);
}

void MidiReset::warning(const char* msg)
{
  printf("Warning: %s\n", msg);
}

void usage()
{
  fprintf(stderr, "midreset adds a sysex reset commands to beginning of a midi file\n");
  fprintf(stderr, "usage: midreset [-xg][-gs][-gm] input.mid output.mid\n");
  fprintf(stderr, "-version\tget program version\n");
  fprintf(stderr, "-xg\tadd XG reset command\n");
  fprintf(stderr, "-gs\tadd GS reset command\n");
  fprintf(stderr, "-gm\tadd GM reset command\n");
  exit(1);
}

int main(int argc, char**argv)
{
  argc--; argv++;
  while (argc > 0 && **argv == '-')
  {
    if (strncmp(*argv, "-version", 2) == 0)
    {
      fprintf(stderr, "%s\n", version);
      argc--; argv++;
      if (argc == 0)
	return 0;
      continue;
    }
    if (strcmp(*argv, "-xg") == 0)
    {
      xg_reset = 1;
      argc--; argv++; continue;
    }
    if (strcmp(*argv, "-gs") == 0)
    {
      gs_reset = 1;
      argc--; argv++; continue;
    }
    if (strcmp(*argv, "-gm") == 0)
    {
      gm_reset = 1;
      argc--; argv++; continue;
    }

    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  if (argc < 2)
    usage();

  input = argv[0];
  output = argv[1];
  if (strcmp(input, output) == 0)
  {
    fprintf(stderr, "cannot convert midi to same file\n");
    return 1;
  }
  MidiReset midi(input);
  if (!midi.getf())
  {
    perror(input);
    return 1;
  }
  write = new MidiWrite(output);
  if (!write)
  {
    fprintf(stderr, "out of memory\n");
    return 1;
  }
  if (!write->getf())
  {
    perror(output);
    return 1;
  }
  argc--; argv++;

  midi.setoutput(write);

  midi.options_ = OPTION_NOCONTROLS;
  if (!midi.run())
    fprintf(stderr, "%s: midi read error at %04lX\n", input, midi.getpos());

  delete write;

  return 0;
}
