// rmi2mid by Gnter Nagler 1995
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef __MSDOS__
#include <mem.h>
#define WRITE_BINARY  "wb"
#define READ_BINARY   "rb"
#else
#define WRITE_BINARY  "w"
#define READ_BINARY   "r"
#endif

char rmifilename[128];
char midfilename[128];

FILE* rmif = 0;
FILE* midf = 0;

long midipos = 0, midilen = 0;

void usage()
{
  printf("rmi2mid file.rmi [file.mid]\n");
}

int readtag(char* tag)
{
  if (fread(tag, 4, 1, rmif) != 1)
  {
    perror(rmifilename);
    return 0;
  }
  tag[4] = '\0';
  return 1;
}

long readtaglen()
{
unsigned char taglenbuf[4];
long len = 0;

  if (fread(taglenbuf, 4, 1, rmif) != 1)
  {
    perror(rmifilename);
    return -1;
  }
  len = taglenbuf[3];
  len = (len << 8) + taglenbuf[2];
  len = (len << 8) + taglenbuf[1];
  len = (len << 8) + taglenbuf[0];
  return len;
}

int isriff()
{
char riff[5];

  fseek(rmif, 0L, SEEK_SET);
  if (!readtag(riff))
    return 0;
  if (strcmp(riff, "RIFF") == 0)
    return 1;
  return 0;
}

int isrmi()
{
char riff[5], midi[5], data[5];
long rifflen;

  fseek(rmif, 0L, SEEK_SET);
  if (!readtag(riff))
    return 0;
  if (strcmp(riff, "RIFF") != 0)
    return 0;
  rifflen = readtaglen();
  if (rifflen <= 0)
    return 0;
  if (!readtag(midi))
    return 0;
  if (strcmp(midi, "RMID") != 0)
    return 0;
  if (!readtag(data))
    return 0;
  if (strcmp(data, "data") != 0)
    return 0;
  midilen = readtaglen();
  if (midilen <= 0)
    return 0;
  midipos = ftell(rmif);
  if (!readtag(midi))
    return 0;
  if (strcmp(midi, "MThd") != 0)
    return 0;
  return 1;
}

static char copybuf[8192];

int rmi2mid()
{
int n;

  if (midilen <= 0)
  {
    if (!isrmi())
      return 0;
  }
  if (midilen <= 0)
    return 0;
  fseek(rmif, midipos, SEEK_SET);
  while (midilen > 0)
  {
    n = sizeof(copybuf);
    if (midilen < n)
      n = (int)midilen;
    if (fread(copybuf, 1, n, rmif) != n)
    {
      perror(rmifilename);
      break;
    }
    if (fwrite(copybuf, 1, n, midf) != n)
    {
      perror(midfilename);
      break;
    }
    midilen -= n;
  }
  return 1;
}

int main(int argc, char**argv)
{
  argc--; argv++;

  if (argc == 0)
  {
    usage();
    return 1;
  }
  strcpy(rmifilename, *argv++); argc--;
  if (argc == 0)
  {
    strcpy(midfilename, rmifilename);
    char* ext = strrchr(midfilename, '.');
    if (ext)
      strcpy(ext, ".mid");
    else
      strcat(midfilename, ".mid");
  }
  else
    strcpy(midfilename, *argv++);

  rmif = fopen(rmifilename, READ_BINARY);
  if (!rmif)
  {
    perror(rmifilename);
    return 1;
  }
  if (!isriff())
  {
    printf("%s is not a riff file\n", rmifilename);
    fclose(rmif);
    return 1;
  }

  if (!isrmi())
  {
    printf("%s is not a riff midi file\n", rmifilename);
    fclose(rmif);
    return 1;
  }
  midf = fopen(midfilename, READ_BINARY);
  if (midf)
  {
    fclose(midf);
    fclose(rmif);
    printf("output filename %s already exists.\n", midfilename);
    return 1;
  }
  midf = fopen(midfilename, WRITE_BINARY);
  if (!midf)
  {
    perror(midfilename);
    fclose(rmif);
    return 1;
  }
  rmi2mid();
  fclose(rmif);
  fclose(midf);
  return 0;
}