%{
// txt2midi written by Gnter Nagler 1995 (gnagler@ihm.tu-graz.ac.at)
static char* version = "txt2midi v1.4 by Gnter Nagler (" __DATE__ ")";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "midiread.hpp"

#ifdef __MSDOS__
#define WRITE_BINARY  "wb"
#define READ_BINARY   "rb"
#define READ_TEXT     "rt"
#define READWRITE_BINARY   "w+b"
#else
#define WRITE_BINARY  "w"
#define READ_BINARY   "r"
#define READ_TEXT     "r"
#define READWRITE_BINARY   "w+"
#endif

void lexstart(char* name);
int yyerror( char* msg );
int yylex();
int yyparse();
void miditext(FILE*f, int tag, char* s);
void midiendtrack(FILE* f);
long log2(long n);
void copybytes(FILE* f, long srcpos, long srclen);
int isactive();
int lastevent = 0xff;
int option_noteoff = 0;
long filesize = 0;
int percent = 0;
int compress = 1;
void needchannel();
void putcode(FILE* f, int code);
void noteon(FILE* f, int note, int vel);
void noteoff(FILE* f, int note, int vel);
void time(long n);

extern char* yytext;
extern long getlinenr();
extern FILE* yyin;
extern FILE* yyout;

unsigned short ntracks = 0;
long ntrackspos = 0;
unsigned long curdelta = 0;
unsigned long sumdelta = 0;
unsigned long clicksperquarter = 192;
unsigned char tact_nom = 4, tact_denom = 4;
unsigned char trackchannel = 0xff;
unsigned char curchannel = 0xff;
unsigned char channel = 0xff;
unsigned char velocyon = 0x7f;
unsigned char velocyoff = 0;
unsigned int  playing = 0xffffU;
unsigned char muting = 1;
unsigned char maketrack = 0;
signed char transpose[16];

typedef struct
{
  char name[80];
  long pos;
  long len;
  long startdelta;  // deltaval before tact
  long enddelta; // deltaval after last tact
  long startsum;
  long endsum;
} MIDIPART;

#define MAXPARTS  10
MIDIPART parts[MAXPARTS];
int partcnt = 0;

%}

%token _TRACK _HEAD _VERSION _END _EVENT _METAEVENT _SYSEVENT
%token _UNIT _TEMPO _BEATS _TRANSPOSE _SIGNEDNUMBER _NUMBER
%token _NOTE _NOTEON _NOTEOFF _TIMES _TEXT _DIV _SEP _ENDTACT
%token _SEQNUMBER _GENERAL_TEXT _COPYRIGHT _TRACKNAME _INSTRUMENT _LYRIC _MARKER _CUEPOINT
%token _SMPTEOFS _TACT _KEY _PROGRAM _CONTROL _VELOCON _VELOCOFF _LOOP
%token _MUTE _SOLO _PART _COPY  _PRINT _BALANCE _REVERB _CHORUS _BRIGHTNESS _HOLD
%token _EXPRESSION _PITCHMODULATION _LEFT _RIGHT _ON _OFF
%token _CHANGE _MODE _SONGPOS _SONGSELECT _COMMON _PITCHBENDRANGE
%token _POLYAFTERTOUCH _AFTERTOUCH _PITCHBEND
%token _HBANK  _LBANK  _BANKA _BANKB _BANKC _BANKD _BANKDRUMS
%token _GMRESET _GSRESET _GSEXIT _XGRESET
%token _PREFIXCHANNEL _PREFIXPORT

%union {
  long longv;
  int intv;
  unsigned char bytev;
  float floatv;
};

%%

start : midisong
  ;

midisong : songoptions midihead songoptions
  {
    memset(transpose, 0, sizeof(transpose));
  }
  miditracks
  {
    fseek(yyout, ntrackspos, SEEK_SET);
    putword(yyout, ntracks);
  }
  ;

songoptions : /* no option */
  | optionlist
  ;

optionlist : optionlist	songoption
  | songoption
  ;

songoption : _MUTE mutechannels
  | _SOLO solochannels
  ;

mutechannels : mutechannels mutechannel
  | mutechannel
  ;

solochannels : solochannels solochannel
  | solochannel
  ;

mutechannel : _NUMBER {
    playing &= ~(1 << (($1.longv-1) & 15));
    muting = 1;
  }
  ;

solochannel : _NUMBER {
    if (muting)
      playing = 0;
    playing |= (1 << (($1.longv-1) & 15));
    muting = 0;
  }
  ;

midihead :
  _HEAD
  {
    open_chunk(yyout, "MThd");
  }
  version
  {
    putword(yyout, (int)$3.longv);  // midi version
    ntrackspos = ftell(yyout);
    putword(yyout, ntracks);
  }
  unit
  {
    clicksperquarter = $5.longv;
    putword(yyout, (unsigned short) $5.longv);
    close_chunk(yyout);
  }
  _END _HEAD
  ;

version : /* default version */ { $$.longv = 1; }
  | _VERSION _NUMBER
    { $$ = $2;
      if ($2.longv != 0 && $2.longv != 1 && $2.longv != 2)
	yyerror("invalid version number: only 0,1,2 are supported");
    }
unit : /* default unit */ { $$.longv = 192; }
  | _UNIT
    _NUMBER
    { $$ = $2; }
  ;

miditracks : miditracks miditrack
  | miditrack
  ;

miditrack :
  _TRACK
  {
    curdelta = sumdelta = 0;
    partcnt = 0;
  }
  channel
  {
    maketrack = 1;
    if (isactive())
    {
      open_chunk(yyout, "MTrk");
      maketrack = 1;
    }
    else
      maketrack = 0;
    lastevent = 0xff;
  }
  opt_midievents
  {
    if (maketrack)
    {
      midiendtrack(yyout);
      close_chunk(yyout);
      ntracks++;
      maketrack = 0;
    }
  }
  _END _TRACK
  ;

channel : { trackchannel = channel = curchannel = 0xff; }
  | '(' _NUMBER ')' {
      curchannel = (unsigned char)(($2.longv-1) & 15);
      if ($2.longv < 1 || $2.longv > 16)
      {
	errormsg("channel out of range 1-16");
	curchannel = 0;
      }
     trackchannel = channel = curchannel;
   }
  ;

opt_midievents : midievents
  |
  ;

midievents : midievents event
  | event
  ;

event : timeevent
  | option
  | _COPY
    _PART
    _TEXT
    {
      copypart(yyout, yytext, &curdelta, &sumdelta);
    }
  | _PART
    _TEXT
    {
      $$.intv = allocpart(yytext);
      if ($$.intv >= 0)
      {
	parts[$$.intv].pos = ftell(yyout);
	parts[$$.intv].startdelta = curdelta;
	parts[$$.intv].startsum = sumdelta;
      }
    }
    midievents
    {
      if ($3.intv >= 0)
      {
	parts[$3.intv].len = ftell(yyout) - parts[$3.intv].pos;
	parts[$3.intv].enddelta = curdelta;
	parts[$3.intv].endsum = sumdelta;
      }
    }
    _END _PART
    _TEXT
    {
      if ($3.intv >= 0)
      {
	if (strcmp(parts[$3.intv].name, yytext) != 0)
	{
	char msg[80];

	  sprintf(msg, "end of part %s not found", parts[$3.intv].name);
	  return yyerror(msg);
	}
      }
    }
  | _LOOP
    _NUMBER
    {
      $$.longv = ftell(yyout);
    }
    {
      $$.longv = curdelta;
    }
    {
      $$.longv = sumdelta;
    }
    midievents
    {

      if (isactive())
      {
      MIDIPART part;
      long i;

	part.name[0] = 0;
	part.pos = $3.longv;
	part.len = ftell(yyout) - $3.longv;
	part.startdelta = $4.longv;
	part.enddelta = curdelta;
	part.startsum = $5.longv;
	part.endsum = sumdelta;

	for (i = 1; i < $2.longv; i++)
	  copypart(yyout, &part, &curdelta, &sumdelta);
      }
    }
    _END _LOOP
  ;

option :
  _VELOCON _NUMBER
    { velocyon = (unsigned char)$2.longv; }
  | _VELOCOFF _NUMBER
    { velocyoff = (unsigned char)$2.longv; }
  | pause sep
  | _PRINT
    {
    if (isactive())
      printf("Channel %d at %ld (%05lXh)\n", channel+1, sumdelta+curdelta, sumdelta+curdelta);
    }
    sep
  | _TRANSPOSE
    transp
    {
      if (curchannel <= 15)
	transpose[curchannel] = (signed char)$2.longv;
    }
    sep
  | sep
  ;

sep : _SEP
  | _ENDTACT
    {
    const epsilon = 0;  // set epsilon to > 0 if small tact errors are allowed
    long tactsize = ((4*clicksperquarter*tact_nom) / tact_denom);
    long tactpos = (sumdelta+curdelta) % tactsize;

      if (tactpos == 0)
	;
      else if (tactpos < epsilon)
      {
	// ok, subtract from curdelta
	if (curdelta >= tactpos)
	  curdelta -= tactpos;
	else
	  ; // not correctable / error in last tact
      }
      else if (tactpos >= tactsize-epsilon)
      {
	// ok, add to curdelta

	curdelta += tactsize-tactpos;
      }
      else
      {
	// error
	yyerror("tact error");
      }
    }
  ;

timeevent :
  {
  int newpercent;

    if (filesize <= 0)
      newpercent = 100;
    else
      newpercent = (int) ((ftell(yyin) * 100) / filesize);
    if (newpercent != percent)
      fprintf(stderr, "%d%%\r", percent = newpercent);
  }
  eventchannel
  midievent
  ;

eventchannel : { channel = curchannel; }
  | '[' _NUMBER ']'
    {
      channel = (int)$2.longv-1;
      if (channel < 0 || channel > 15)
      {
	errormsg("event channel out of range 1-16");
	channel = 0;
      }
      if (trackchannel == 0xff || trackchannel == channel)
	curchannel = channel;
      else /* invalid channel */
      {
	yyerror("multiple channels used in single channel track");
	channel = curchannel;
      }
    }
  ;

midievent :
    _SEQNUMBER
    _NUMBER
    {
      if (isactive())
      {
	putcode(yyout, 0xff);
	putbyte(yyout, 0);
	putbyte(yyout, 2);
	putword(yyout, (unsigned short)$2.longv);
      }
    }
  | _GENERAL_TEXT _TEXT
    {
      miditext(yyout, 1, yytext);
    }
  | _COPYRIGHT _TEXT
    {
      miditext(yyout, 2, yytext);
    }
  | _TRACKNAME _TEXT
    {
      miditext(yyout, 3, yytext);
    }
  | _INSTRUMENT _TEXT
    {
      miditext(yyout, 4, yytext);
    }
  | _LYRIC _TEXT
    {
      miditext(yyout, 5, yytext);
    }
  | _MARKER _TEXT
    {
      miditext(yyout, 6, yytext);
    }
  | _CUEPOINT _TEXT
    {
      miditext(yyout, 7, yytext);
    }
  | _SMPTEOFS _NUMBER ":" _NUMBER ":" _NUMBER ":" _NUMBER ":" _NUMBER
    {
      if (isactive())
      {
	putcode(yyout, 0xff);
	putbyte(yyout, 0x54);
	putbyte(yyout, 5); // length
	putbyte(yyout, (unsigned char)$2.longv);
	putbyte(yyout, (unsigned char)$4.longv);
	putbyte(yyout, (unsigned char)$6.longv);
	putbyte(yyout, (unsigned char)$8.longv);
	putbyte(yyout, (unsigned char)$10.longv);
      }
    }
  | _TACT _NUMBER _DIV _NUMBER _NUMBER _NUMBER
    {
      if (isactive())
      {
	putcode(yyout, 0xff);
	putbyte(yyout, 0x58);
	putbyte(yyout, 4); // length
	putbyte(yyout, tact_nom = (unsigned char)$2.longv);
	putbyte(yyout, (unsigned char)log2(tact_denom = (unsigned char)$4.longv));
	putbyte(yyout, (unsigned char)$5.longv);
	putbyte(yyout, (unsigned char)$6.longv);
      }
    }
  | _PREFIXCHANNEL _NUMBER
    {
      if (isactive())
      {
	putcode(yyout, 0xff);
	putbyte(yyout, 0x20);
	putbyte(yyout, 1);
	putbyte(yyout, (unsigned char)(($2.longv-1) & 15));
      }
    }
  | _PREFIXPORT _NUMBER
    {
      if (isactive())
      {
	putcode(yyout, 0xff);
	putbyte(yyout, 0x21);
	putbyte(yyout, 1);
	putbyte(yyout, (unsigned char)($2.longv & 15));
      }
    }
  | _TEMPO _NUMBER
    {
      if (isactive())
      {
	putcode(yyout, 0xff);
	putbyte(yyout, 0x51);
	putbyte(yyout, 3);
	puttri(yyout, $2.longv);
      }
    }
  | _BEATS
    {
      if (isactive())
      {
	putcode(yyout, 0xff);
	putbyte(yyout, 0x51);
	putbyte(yyout, 3);
	if ($1.floatv < 1.0)
	{
	  yyerror("beats value is too small (< 1.0)");
	  $1.floatv = 120.0; // default tempo
	}
	puttri(yyout, (long)(60000000.0 / $1.floatv));
      }
    }
  | _KEY _TEXT
    {
    int signature = 0, isminor = 0;
    char* s = yytext;

      if (*s == 'C')
      {
	signature = 0;
	s++;
      }
      else if (*s >= '1' && *s <= '7' && strchr("b#", s[1]) != 0)
      {
	signature = *s++ - '0';
	if (*s++ == 'b')
	  signature = -signature;
      }
      else
      {
	yyerror("invalid key signature should be: C or 1-7b 1-7#");
	*s = 0;
      }
      while (isspace(*s))
	s++;
      if (strncmp(s, "major", 3) == 0 || strncmp(s, "Major", 3) == 0)
	isminor = 0;
      else if (strncmp(s, "minor", 3) == 0 || strncmp(s, "Minor", 3) == 0)
	isminor = 1;
      else if (*s)
	yyerror("illegal key signature: missing maj or min");

      if (isactive())
      {
	putcode(yyout, 0xff);
	putbyte(yyout, 0x59);
	putbyte(yyout, 2);
	putbyte(yyout, (unsigned char)signature);
	putbyte(yyout, (unsigned char)isminor);
      }
    }
  | _EVENT
    {
      time(curdelta);
    }
    bytes
    _END _EVENT
  | _METAEVENT
    _NUMBER
    '(' _NUMBER ')'
    { if (isactive())
      {
	putcode(yyout, 0xff);
	if ($2.longv < 0 || $2.longv >= 128)
	  yyerror("invalid metaevent (should be in range 0-127)");
	putbyte(yyout, (unsigned char)$2.longv);
	putdelta(yyout, $4.longv);
      }
    }
    bytes
    _END _METAEVENT
    {
      if ($7.longv != $4.longv)
	yyerror("invalid length specified in metaevent");
    }
  | _SYSEVENT
    '(' _NUMBER ')'
    { if (isactive())
	putcode(yyout, 0xf0);
    }
    bytes
    _END _SYSEVENT
    { if (isactive())
       putbyte(yyout, 0xf7);
      if ($6.longv != $3.longv)
	yyerror("invalid length specified in sysevent");
    }
  | _GMRESET
    { if (isactive())
      {
	// F0 05 7E 7F 09 01 F7

	putcode(yyout, 0xf0);
	putbyte(yyout, 0x05);
	putbyte(yyout, 0x7e);
	putbyte(yyout, 0x7f);
	putbyte(yyout, 0x09);
	putbyte(yyout, 0x01);
	putbyte(yyout, 0xf7);
      }
    }
  | _GSRESET
    { if (isactive())
      {
	// F0 0A 41 10 42 12 40 00 7F 00 41 F7

	putcode(yyout, 0xf0);
	putbyte(yyout, 0x0A);
	putbyte(yyout, 0x41);
	putbyte(yyout, 0x10);
	putbyte(yyout, 0x42);
	putbyte(yyout, 0x12);
	putbyte(yyout, 0x40);
	putbyte(yyout, 0x00);
	putbyte(yyout, 0x7f);
	putbyte(yyout, 0x00);
	putbyte(yyout, 0x41);
	putbyte(yyout, 0xf7);
      }
    }
  | _XGRESET
    { if (isactive())
      {
	// F0 08 43 10 4C 00 00 7E 00 F7

	putcode(yyout, 0xf0);
	putbyte(yyout, 0x08);
	putbyte(yyout, 0x43);
	putbyte(yyout, 0x10);
	putbyte(yyout, 0x4C);
	putbyte(yyout, 0x00);
	putbyte(yyout, 0x00);
	putbyte(yyout, 0x7e);
	putbyte(yyout, 0x00);
	putbyte(yyout, 0xf7);
      }
    }
  | _GSEXIT
    { if (isactive())
      {
	// F0 0A 41 10 42 12 40 00 7F 7F 42 F7

	putcode(yyout, 0xf0);
	putbyte(yyout, 0x0A);
	putbyte(yyout, 0x41);
	putbyte(yyout, 0x10);
	putbyte(yyout, 0x42);
	putbyte(yyout, 0x12);
	putbyte(yyout, 0x40);
	putbyte(yyout, 0x00);
	putbyte(yyout, 0x7f);
	putbyte(yyout, 0x7f);
	putbyte(yyout, 0x42);
	putbyte(yyout, 0xf7);
      }
    }
  | _PROGRAM
    program
    {
      if (isactive())
      {
	needchannel();
	putcode(yyout, 0xC0+channel);
	putbyte(yyout, (unsigned char)$2.longv);
      }
    }
  | _CONTROL
    _NUMBER
    _NUMBER
    {
      if (isactive())
      {
	needchannel();
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, (unsigned char)$2.longv);
	putbyte(yyout, (unsigned char)$3.longv);
      }
    }
  | _HBANK
    _NUMBER
     {
       if (isactive())
       {
	needchannel();
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, 0x00);
	putbyte(yyout, (unsigned char)$2.longv);
       }
     }
  | _LBANK
    _NUMBER
     {
       if (isactive())
       {
	needchannel();
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, 0x20);
	putbyte(yyout, (unsigned char)$2.longv);
       }
     }
  | bank
     {
       if (isactive())
       {
	needchannel();
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, 0x00);
	putbyte(yyout, (int)(($1.longv >> 8) & 0xff));
	putbyte(yyout, 0x0); // time
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, 0x20);
	putbyte(yyout, (int)($1.longv & 0xff));
       }
     }
  | _BALANCE
    balance
    {
      if (isactive())
      {
	needchannel();
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, 10);
	putbyte(yyout, (unsigned char)$2.longv);
      }
    }
  | _HOLD
    onoff
    {
      if (isactive())
      {
	needchannel();
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, 0x40);
	putbyte(yyout, (unsigned char)$2.longv);
      }
    }
  | _REVERB
    _NUMBER
    {
      if (isactive())
      {
	needchannel();
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, 0x5b);
	putbyte(yyout, (unsigned char)$2.longv);
      }
    }
  | _CHORUS
    _NUMBER
    {
      if (isactive())
      {
	needchannel();
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, 0x5d);
	putbyte(yyout, (unsigned char)$2.longv);
      }
    }
  | _BRIGHTNESS
    _NUMBER
    {
      if (isactive())
      {
	needchannel();
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, 0x4a);
	putbyte(yyout, (unsigned char)$2.longv);
      }
    }
  | _EXPRESSION
    _NUMBER
    {
      if (isactive())
      {
	needchannel();
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, 11);
	putbyte(yyout, (unsigned char)$2.longv);
      }
    }
  | _PITCHMODULATION
    _NUMBER
    {
      if (isactive())
      {
	needchannel();
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, 1);
	putbyte(yyout, (unsigned char)$2.longv);
      }
    }
  | _CHANGE
    _NUMBER
    {
      if (isactive())
      {
	needchannel();
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, $1.bytev);
	putbyte(yyout, (unsigned char)$2.longv);
      }
    }
  | _PITCHBENDRANGE
    _NUMBER
    {
      if (isactive())
      {
	needchannel();
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, 0x65); // highRPN 0
	putbyte(yyout, 0);
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, 0x64); // lowRPN 0
	putbyte(yyout, 0);
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, 6); // data
	putbyte(yyout, (unsigned char)$2.longv);
      }
    }
  | _MODE
    {
      if (isactive())
      {
	needchannel();
	putcode(yyout, 0xB0+channel);
	putbyte(yyout, (unsigned char)($1.longv & 0xff));
	putbyte(yyout, (unsigned char)(($1.longv >> 8) & 0xff));
      }
    }
  | _SONGPOS
    _NUMBER
    _NUMBER
    {
      if (isactive())
      {
	putcode(yyout, 0xF2);
	putbyte(yyout, (unsigned char)($2.longv & 127));
	putbyte(yyout, (unsigned char)($3.longv & 127));
      }
    }
  | _SONGSELECT
    _NUMBER
    {
      if (isactive())
      {
	putcode(yyout, 0xF3);
	putbyte(yyout, (unsigned char)(($2.longv & 127)));
      }
    }
  | _COMMON
    {
      if (isactive())
      {
	putcode(yyout, $1.bytev);
      }
    }
  | _POLYAFTERTOUCH
    note
    _NUMBER
    {
      if (isactive())
      {
	putcode(yyout, 0xA0+channel);
	putbyte(yyout, (unsigned char)$2.longv);
	putbyte(yyout, (unsigned char)($3.longv & 127));
      }
    }
  | _AFTERTOUCH
    _NUMBER
    {
      if (isactive())
      {
	putcode(yyout, 0xD0+channel);
	putbyte(yyout, (unsigned char)($2.longv & 127));
      }
    }
  | _PITCHBEND
    _NUMBER
    {
      if (isactive())
      {
	putcode(yyout, 0xE0+channel);
	putbyte(yyout, (unsigned char)(($2.longv & 127)));
	putbyte(yyout, (unsigned char)((($2.longv >> 7) & 127)));
      }
    }
  | _NOTE
    duration
    velocyon
    {
      if (isactive())
      {
	noteon(yyout, (unsigned char)$1.longv, (unsigned char)$3.longv);
	time($2.longv);
	noteoff(yyout, (unsigned char)$1.longv, 0);
      }
    }
    sep
  | _NOTEON
    note
    velocyon
    {
      if (isactive())
      {
	putcode(yyout, 0x90+channel);
	putbyte(yyout, (unsigned char)$2.longv);
	putbyte(yyout, (unsigned char)$3.longv);
      }
    }
    sep
  | _NOTEOFF
    note
    velocyoff
    {
      if (isactive())
	noteoff(yyout, (unsigned char)$2.longv, (unsigned char)$3.longv);
    }
    sep
  | _SIGNEDNUMBER
    {
      if (isactive())
      {
	if ($1.longv < 0) // note off
	  noteoff(yyout, (unsigned char)(-$1.longv), velocyoff);
	else // note on
	  noteon(yyout, (unsigned char)$1.longv, velocyon);
      }
    }
  ;

note : _NOTE { $$.longv = $1.longv + transpose[curchannel]; if ($$.longv < 0) return yyerror("invalid note"); }
  | _NUMBER { $$.longv = $1.longv + transpose[curchannel]; if ($$.longv < 0) return yyerror("invalid note"); }
  ;

bank : _BANKA { $$.longv = 0x3800; }
  | _BANKB { $$.longv = 0x3800; }
  | _BANKC { $$.longv = 0x0001; }
  | _BANKD { $$.longv = 0x0001; }
  | _BANKDRUMS { $$.longv = 0x3E00; }
  ;

velocyon : _NUMBER
  | { $$.longv = velocyon; }
  ;

velocyoff : _NUMBER
  | { $$.longv = velocyoff; }
  ;

balance : _LEFT { $$.longv = 0; }
  | _RIGHT { $$.longv = 127; }
  | _NUMBER
  ;

onoff : _ON { $$.longv = 127; }
  | _OFF { $$.longv = 0; }
  | _NUMBER
  ;

pause : duration
  {
    curdelta += $1.longv;
  }
  ;

duration : _NUMBER
  | _NUMBER _DIV _NUMBER { $$.longv = $1.longv * 4 * clicksperquarter / $3.longv ; };
  ;

transp :
  _SIGNEDNUMBER
  | _NUMBER
  ;

program : _NUMBER
  ;

bytes : bytes byte { $$.longv = $1.longv + 1; }
  | bytes string  { $$.longv = $1.longv + $2.longv; }
  | byte { $$.longv = 1; }
  | string
  ;

byte :
  _NUMBER
  {
    if (isactive())
    {
      putbyte(yyout, (unsigned char)$1.longv);
    }
  }
  ;

string :
  _TEXT
  {
    if (isactive())
    {
    char* s = yytext;
    int len = 0;

      while (*s)
      {
	putbyte(yyout, *s++);
	len++;
      }
      $$.longv = len;
    }
    else
      $$.longv = 0;
  }
  ;

/*
tokens : tokens token
  | token
  ;

token : tok { printf("%s\n", yytext); }
  ;

tok : _TRACK | _HEAD | _END | _EVENT | _SYSEVENT |
  _TEMPO | _NUMBER | _NOTE | _TIMES | _TEXT | _DIV | _SEP | _ENDTACT |
  _GENERAL_TEXT | _COPYRIGHT | _TRACKNAME | _INSTRUMENT | _LYRIC |
  _CHANNEL | _TACT | _PROGRAM | _CONTROL | _VELOCON | _VELOCOFF
  ;
*/

%%

char* infile = 0;
char* outfile = 0;

int main(int argc, char**argv)
{
  argc--; argv++;

  while (argc > 0 && **argv == '-')
  {
    if (strncmp(*argv, "-version", 2) == 0)
    {
      fprintf(stderr, "%s\n", version);
      argc--; argv++;
      if (argc == 0)
	return 0;
      continue;
    }
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
  }
  if (argc < 2)
  {
    fprintf(stderr, "usage: txt2midi [-version] textfile.txt midifile.mid\n");
    return 1;
  }

  yyin = fopen(infile = *argv++, READ_TEXT); argc--;
#ifdef __MSDOS__
    strlwr(infile);
#endif
  if (!yyin)
  {
    perror(infile);
    return 1;
  }
  fseek(yyin, 0L, SEEK_END);
  filesize = ftell(yyin);
  fseek(yyin, 0L, SEEK_SET);

  if (argc == 0)
    yyout = 0;
  else
  {
    outfile = *argv++; argc--;
#ifdef __MSDOS__
    strlwr(outfile);
#endif
    if (strcmp(infile, outfile) == 0)
    {
      yyout = 0;
      fprintf(stderr, "input and output files are identical!\n");
    }
    else
    {
      yyout = fopen(outfile, READ_BINARY);
      if (yyout)
      {
      MIDIHEAD h;
      int ismidi;

	// test if it is a midi

	ismidi = get_midihead(yyout, &h);
	fclose(yyout);
	if (!ismidi)
	{
	  fprintf(stderr, "cannot overwrite existing file %s\n", outfile);
	  return 1;
	}
      }
      yyout = fopen(outfile, READWRITE_BINARY);
      if (!yyout)
      {
	perror(outfile);
	return 1;
      }
    }
  }
  lexstart(infile);
  yyparse();

  if (percent < 100 && ftell(yyin) >= filesize)
    fprintf(stderr, "%d\r", percent = 100);
  if (yyin)
    fclose(yyin);
  if (yyout)
    fclose(yyout);
  return 0;
}

void errormsg(char* msg)
{
  fprintf(stderr, "Line %ld at '%s': %s\n", getlinenr(), yytext, msg );
  if (strncmp(msg, "invalid program", 15) == 0)
  {
  int i;

    fprintf(stderr, "valid program names are: \n");
    fprintf(stderr, "A11-D88, Dr1-Dr8, 0-127\n");
    for (i = 0; i < 128; i++)
    {
      fprintf(stderr, "%s", progname(i, 0));
      if (i % 8 != 7)
	fputc(' ', stderr);
      else
	fputc('\n', stderr);
    }
  }
  if (yyout)
  {
    fclose(yyout);
    if (outfile && *outfile)
      unlink(outfile);
  }
}

int yyerror( char* msg )
{
  errormsg(msg);
  exit(1);
  return 1;
}

void miditext( FILE* f, int tag, char* s )
{
int len;

  if (isactive())
  {
    putcode(f, 0xff);
    putbyte(f, tag);
    putdelta(f, len = strlen(s));
    if (len > 0)
      fwrite(s, 1, len, f);
  }
}

void midiendtrack(FILE* f)
{
  putcode(f, 0xff);
  putbyte(f, 0x2f);
  putdelta(f, 0); // end of track
}

long log2(long n)
{
long l = 0;

  while (n > 1)
  {
    n /= 2;
    l++;
  }
  return l;
}

void copybytes(FILE* f, long srcpos, long srclen)
{
long dstpos = ftell(f);
char s[512];
int n,r = 0;

  while (srclen > 0)
  {
    n = 512;
    if (srclen < n)
      n = (int)srclen;
    fseek(f, srcpos, SEEK_SET);
    r = fread(s, 1, n, f);
    if (r <= 0)
    {
      yyerror("cannot read midi file");
      return;
    }
    srcpos += r;
    srclen -= r;
    fseek(f, dstpos, SEEK_SET);
    n = fwrite(s, 1, r, f);
    dstpos += n;
  }
}

int isactive()
{
  if (!maketrack)
    return 0;
  if (trackchannel <= 15)   // single channel tracks
    return playing & (1 << trackchannel);
  if (trackchannel == 0xff) // multi channel track
  {
    if (curchannel == 0xff) // not used yet
      return 1;
    return playing & (1 << curchannel);
  }
  return 0;
}

void copypart(FILE* f, MIDIPART* part, unsigned long *curdelta, unsigned long *sumdelta)
{
long d, pos, ppos, len;

  pos = ftell(f);

  if (pos > part->pos)
  {
    fseek(yyout, part->pos, SEEK_SET);
    d = getdelta(yyout);
    ppos = ftell(yyout);
    len = part->len - (ppos-part->pos);
    fseek(yyout, pos, SEEK_SET);
    if (len > 0)
    {
      putdelta(f, *curdelta + d - part->startdelta);
      *sumdelta += part->endsum-part->startsum + *curdelta - part->startdelta;
      *curdelta = part->enddelta;
      copybytes(yyout, ppos, len);
    }
  }
}

void initpart(MIDIPART& p, char* name)
{
  strcpy(p.name, name);
  p.pos = -1;
  p.startdelta = p.enddelta = -1;
  p.startsum = p.endsum = 0;
  p.len = 0;
}

int findpart(char* name)
{
int i;

  for (i = 0; i < partcnt; i++)
    if (strcmp(parts[i].name, name) == 0)
      return i;
  return -1;
}

int allocpart(char* name)
{
int i;

  i = findpart(name);
  if (i >= 0)
  {
  char msg[80];

    sprintf(msg, "duplicate part name %s", name);
    yyerror(msg);
    return -1;
  }
  for (i = 0; i < partcnt; i++)
    if (parts[i].name[0] == 0)
      break;

  if (i == partcnt)
  {
    if (partcnt < MAXPARTS)
      i = partcnt++;
    else
      return -1;
  }

  initpart(parts[i], name);
  return i;
}

void copypart(FILE* f, char* partname, unsigned long *curdelta, unsigned long *sumdelta)
{
int i;

  i = findpart(partname);
  if (i >= 0)
  {
    copypart(f, &parts[i], curdelta, sumdelta);
  }
  else
  {
  char msg[80];

    sprintf(msg, "part not defined %s", partname);
    yyerror(msg);
  }
}

void putcode(FILE* f, int code)
{
int put;

  time(curdelta);

  if (compress)
    put = !(code == lastevent && code <= 0x9f);
  else
    put = 1;
  if (put)
    putbyte(f, code);
  lastevent = code;
}

void noteoff(FILE* f, int note, int vel)
{
  if (vel != 0 || option_noteoff)
    putcode(f, 0x80+channel);
  else
    putcode(f, 0x90+channel);
  putbyte(f, (unsigned char)note);
  putbyte(f, (unsigned char)vel);
}


void noteon(FILE* f, int note, int vel)
{
  putcode(f, 0x90+channel);
  putbyte(f, (unsigned char)note);
  putbyte(f, (unsigned char)vel);
}

void time(long n)
{
  putdelta(yyout, n);
  sumdelta += n;
  curdelta = 0;
}

void needchannel()
{
static int hint = 0;

  if (channel >= 0 && channel <= 15)
    return;
  errormsg("midi command missing channel number (1-16), using default channel 1");
  channel = 0;
  if (!hint)
  {
    hint = 1;
    printf("Info: set track channel using `mtrk(channel)` or set event channel using [channel]\n");
  }
}
